#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

'''
This module implements some basic matrix classes
'''
from math import log, sqrt, cos, sin, atan2, asin, acos, pi
from sys import float_info


class CannotRowReduce(ValueError): pass
class MatrixNotInvertable(ValueError): pass
class QuaternionNotInvertable(ValueError): pass


def are_vectors_equal(vector_0, vector_1, use_double_rounding=False,
                      round_adjust=0):
    mantissa_len = (53 if use_double_rounding else 23)
    for val_0, val_1 in zip(vector_0, vector_1):
        # take into account rounding errors for 32bit floats
        val = max(abs(val_0), abs(val_1))
        delta_max = 2**(
            int(log(val + float_info.epsilon, 2)) -
            mantissa_len) + abs(round_adjust)
        if abs(val_1 - val_0) > delta_max:
            return False
    return True


def vertex_cross_product(v0, v1, v2):
    return cross_product(line_from_verts(v0, v1),
                         line_from_verts(v0, v2))


def cross_product(ray_a, ray_b):
    return ((ray_a[1]*ray_b[2] - ray_a[2]*ray_b[1],
             ray_a[2]*ray_b[0] - ray_a[0]*ray_b[2],
             ray_a[0]*ray_b[1] - ray_a[1]*ray_b[0]))


def dot_product(v0, v1):
    return sum(a*b for a, b in zip(v0, v1))


def line_from_verts(v0, v1):
    return tuple(b - a for a, b in zip(v0, v1))


def euler_2d_to_vector_3d(y, p):
    '''Angles are expected to be in radians.'''
    return (cos(y) * cos(p),
            sin(y) * cos(p),
            sin(p))


def euler_to_quaternion(y, p, r):
    '''Angles are expected to be in radians.'''
    c0, c1, c2 = cos(y / 2), cos(p / 2), cos(r / 2)
    s0, s1, s2 = sin(y / 2), sin(p / 2), sin(r / 2)
    return (s0*s1*c2 + c0*c1*s2, s0*c1*c2 + c0*s1*s2,
            c0*s1*c2 - s0*c1*s2, c0*c1*c2 - s0*s1*s2)


def quaternion_to_euler(i, j, k, w):
    '''Angles returned are in radians.'''
    p_sin = 2*(i * j + k * w)
    # check for singularities at north and south poles
    if p_sin > 0.999999999999:
        return 2 * atan2(i, w),   pi / 2, 0
    elif p_sin < -0.999999999999:
        return -2 * atan2(i, w), -pi / 2, 0
    else:
        y = atan2(2*(j*w - i*k), 1 - 2*(j**2 + k**2))
        p = asin(p_sin)
        r = atan2(2*(i*w - j*k), 1 - 2*(i**2 + k**2))
        return y, p, r


def axis_angle_to_quaternion(x, y, z, a):
    '''Angle is expected to be in radians.'''
    a /= 2
    sa = sin(a)
    return x * sa, y * sa, z * sa, cos(a)


def quaternion_to_axis_angle(i, j, k, w):
    ray_len = sqrt(i**2 + j**2 + k**2 + w**2)
    i /= ray_len
    j /= ray_len
    k /= ray_len
    w /= ray_len

    length = sqrt(1 - w**2)
    if length == 0.0:
        return i, j, k, 0
    return i / length, j / length, k / length, 2 * acos(w)


def quaternion_to_matrix(i, j, k, w):
    return Matrix([
        (2*(0.5 - j*j - k*k),   2*(i*j + k*w),         2*(i*k - j*w)),
        (2*(i*j - k*w),         2*(0.5 - k*k - i*i),   2*(j*k + i*w)),
        (2*(i*k + j*w),         2*(j*k - i*w),         2*(0.5 - i*i - j*j)),
    ])


def matrix_to_quaternion(matrix):
    m00, m10, m20 = matrix[0]
    m01, m11, m21 = matrix[1]
    m02, m12, m22 = matrix[2]
    tr = m00 + m11 + m22

    if tr > 0:
        s = sqrt(tr+1.0) * 2
        i = (m21 - m12) / s
        j = (m02 - m20) / s
        k = (m10 - m01) / s
        w = 0.25 * s
    elif m00 > m11 and m00 > m22:
        s = sqrt(1.0 + m00 - m11 - m22) * 2
        i = 0.25 * s
        j = (m01 + m10) / s
        k = (m02 + m20) / s
        w = (m21 - m12) / s
    elif m11 > m22:
        s = sqrt(1.0 + m11 - m00 - m22) * 2
        i = (m01 + m10) / s
        j = 0.25 * s
        k = (m12 + m21) / s
        w = (m02 - m20) / s
    else:
        s = sqrt(1.0 + m22 - m00 - m11) * 2
        i = (m02 + m20) / s
        j = (m12 + m21) / s
        k = 0.25 * s
        w = (m10 - m01) / s

    return i, j, k, w


def multiply_quaternions(q0, q1):
    i =  q0[0] * q1[3] + q0[1] * q1[2] - q0[2] * q1[1] + q0[3] * q1[0]
    j = -q0[0] * q1[2] + q0[1] * q1[3] + q0[2] * q1[0] + q0[3] * q1[1]
    k =  q0[0] * q1[1] - q0[1] * q1[0] + q0[2] * q1[3] + q0[3] * q1[2]
    w = -q0[0] * q1[0] - q0[1] * q1[1] - q0[2] * q1[2] + q0[3] * q1[3]
    div = i**2 + j**2 + k**2 + w**2
    if not div:
        # not sure if we should raise an error.
        # this multiplication makes no sense.
        i = j = k = 0.0
        w = 1.0
    else:
        div = sqrt(div)
        i /= div
        j /= div
        k /= div
        w /= div

    return type(q0)((i, j, k, w))


# shorthand names
matrix_to_quat = matrix_to_quaternion
quat_to_matrix = quaternion_to_matrix
euler_to_quat = euler_to_quaternion
quat_to_euler = quaternion_to_euler
axis_angle_to_quat = axis_angle_to_quaternion
quat_to_axis_angle = quaternion_to_axis_angle
multiply_quats = multiply_quaternions


def clip_angle_to_bounds(angle, step=pi):
    angle -= step * 2 * int(angle / (2 * step))

    if angle < -step:
        return angle + step * 2
    elif angle > step:
        return angle - step * 2
    return angle


def lerp_blend_vectors(v0, v1, ratio):
    r1 = max(0.0, min(1.0, ratio))
    r0 = 1.0 - r1
    return [a*r0 + b*r1 for a, b in zip(v0, v1)]


def nlerp_blend_quaternions(q0, q1, ratio):
    r1 = max(0.0, min(1.0, ratio))
    r0 = 1.0 - ratio

    i0, j0, k0, w0 = q0
    i1, j1, k1, w1 = q1

    cos_half_theta = i0*i1 + j0*j1 + k0*k1 + w0*w1
    if cos_half_theta < 0:
        # need to change the vector rotations to be 2pi - rot
        r1 = -r1

    return [i0*r0 + i1*r1, j0*r0 + j1*r1, k0*r0 + k1*r1, w0*r0 + w1*r1]


def slerp_blend_quaternions(q0, q1, ratio):
    ratio = max(0.0, min(1.0, ratio))

    i0, j0, k0, w0 = q0
    i1, j1, k1, w1 = q1

    cos_half_theta = i0*i1 + j0*j1 + k0*k1 + w0*w1
    half_theta = acos(cos_half_theta) if abs(cos_half_theta) < 1.0 else 0.0

    if cos_half_theta < 0:
        # need to change the vector rotations to be 2pi - rot
        cos_half_theta = -cos_half_theta
        i1 = -i1
        j1 = -j1
        k1 = -k1
        w1 = -w1

    # angle is not well defined in floating point at this point
    if cos_half_theta > 0.9999999:
        r0 = 1.0 - ratio
        r1 = ratio
    else:
        sin_half_theta = sqrt(max(1 - cos_half_theta**2, 0))
        r0 = sin((1.0 - ratio) * half_theta) / sin_half_theta
        r1 = sin(ratio * half_theta) / sin_half_theta

    return [i0*r0 + i1*r1, j0*r0 + j1*r1, k0*r0 + k1*r1, w0*r0 + w1*r1]


class FixedLengthList(list):
    __slots__ = ()
    def append(self, val): raise NotImplementedError
    def extend(self, vals): raise NotImplementedError
    def insert(self, index, val): raise NotImplementedError
    def pop(self): raise NotImplementedError
    def __delitem__(self): raise NotImplementedError
    def __setitem__(self, index, val):
        if isinstance(index, slice):
            start, stop, step = index.indices(len(self))
            if start > stop:
                start, stop = stop, start
            if start == stop:
                return
            elif step < 0:
                step = -step

            slice_size = (stop - start) // step

            if slice_size != len(val):
                raise ValueError(("attempt to assign sequence of size %s to "
                                  "slice of size %s") % (len(val), slice_size))

        list.__setitem__(self, index, val)


class Vector(list):
    __slots__ = ()
    '''Implements the minimal methods required for messing with matrix rows'''
    def __neg__(self):
        return type(self)(-x for x in self)
    def __add__(self, other):
        new = type(self)(self)
        for i in range(len(other)): new[i] += other[i]
        return new
    def __sub__(self, other):
        new = type(self)(self)
        for i in range(len(other)): new[i] -= other[i]
        return new
    def __mul__(self, other):
        if isinstance(other, type(self)):
            return sum(self[i]*other[i] for i in range(len(self)))
        new = type(self)(self)
        for i in range(len(self)): new[i] *= other
        return new
    def __truediv__(self, other):
        if isinstance(other, type(self)):
            return cross_product(self, other)
        new = type(self)(self)
        for i in range(len(self)): new[i] /= other
        return new
    def __eq__(self, other):
        return are_vectors_equal(self, other)
    def __iadd__(self, other):
        for i in range(len(other)): self[i] += other[i]
        return self
    def __isub__(self, other):
        for i in range(len(other)): self[i] -= other[i]
        return self
    def __imul__(self, other):
        if isinstance(other, type(self)):
            raise NotImplementedError
        for i in range(len(self)): self[i] *= other
        return self
    def __itruediv__(self, other):
        if isinstance(other, type(self)):
            raise NotImplementedError
        for i in range(len(self)): self[i] /= other
        return self

    __radd__ = __add__
    __rsub__ = __sub__
    __rmul__ = __mul__
    __rtruediv__ = __truediv__


class Point(Vector):
    __slots__ = ()


class Ray(Vector):
    __slots__ = ()
    @property
    def is_zero(self): return not sum(bool(val) for val in self)
    @property
    def normalized(self):
        ray = type(self)(self)
        ray.normalize()
        return ray
    @property
    def magnitude_sq(self): return sum(v**2 for v in self)
    @property
    def magnitude(self): return sqrt(sum(v**2 for v in self))
    mag_sq = magnitude_sq
    mag = magnitude

    @classmethod
    def cross(cls, v0, v1):
        assert len(v0) >= 3
        assert len(v1) >= 3
        return Ray(cross_product(v0, v1))
    @classmethod
    def dot(cls, v0, v1):
        assert len(v0) == len(v1)
        return dot_product(v0, v1)

    def normalize(self):
        div = self.magnitude
        if div:
            for i in range(len(self)): self[i] /= div


class Quaternion(FixedLengthList, Ray):
    __slots__ = ()
    def __init__(self, initializer=(0, 0, 0, 1)):
        assert len(initializer) == 4
        list.__init__(self, initializer)
    def __mul__(self, other):
        if isinstance(other, Quaternion):
            new = Quaternion(multiply_quaternions(self, other))
        else:
            new = Ray(self)
            for i in range(len(self)): new[i] *= other
        return new
    def __imul__(self, other):
        if isinstance(other, Quaternion):
            return Quaternion(multiply_quaternions(self, other))
        for i in range(len(self)): self[i] *= other
        return self
    __rmul__ = __mul__

    @property
    def i(self): return self[0]
    @i.setter
    def i(self, new_val): self[0] = float(new_val)
    @property
    def j(self): return self[1]
    @j.setter
    def j(self, new_val): self[1] = float(new_val)
    @property
    def k(self): return self[2]
    @k.setter
    def k(self, new_val): self[2] = float(new_val)
    @property
    def w(self): return self[3]
    @w.setter
    def w(self, new_val): self[3] = float(new_val)

    @property
    def to_euler(self):
        return Vector(quaternion_to_euler(*self))
    @property
    def to_axis_angle(self):
        return Vector(quaternion_to_axis_angle(*self))
    @property
    def to_matrix(self):
        return quaternion_to_matrix(*self)
    @property
    def inverse(self):
        copy = Quaternion(self)
        copy.invert()
        return copy

    def invert(self):
        div = self.i**2 + self.j**2 + self.k**2 + self.w**2
        if not div:
            raise QuaternionNotInvertable("%s is not invertable." % self)

        div = sqrt(div)
        self.i = -self.i / div
        self.j = -self.j / div
        self.k = -self.k / div
        self.w =  self.w / div


class MatrixRow(FixedLengthList, Vector):
    __slots__ = ()


class Matrix(list):
    width = 1
    height = 1

    def __init__(self, matrix=None, width=1, height=1, identity=False):
        if matrix is None:
            self.width = width
            self.height = height
            list.__init__(self, (MatrixRow((0,)*width) for i in range(height)))

            if identity and width <= height:
                # place the identity matrix into the inverse
                for i in range(self.width):
                    self[i][i] = 1.0
            return

        matrix_rows = []
        self.height = max(1, len(matrix))
        self.width = -1
        for row in matrix:
            if not hasattr(row, '__iter__'):
                row = [row]
            self.width = max(self.width, len(row))
            assert self.width and len(row) == self.width
            matrix_rows.append(MatrixRow(row[:]))
        list.__init__(self, matrix_rows)

    def __setitem__(self, index, new_row):
        assert len(new_row) == self.width
        list.__setitem__(self, index, MatrixRow(new_row))

    def __delitem__(self, index):
        self[index][:] = (0,)*self.width

    def __str__(self):
        matrix_str = "Matrix([\n%s])"
        insert_str = ''
        for row in self:
            insert_str += '%s,\n' % (row,)
        return matrix_str % insert_str

    def __neg__(self):
        return Matrix([-row for row in self])

    def __add__(self, other):
        assert isinstance(other, Matrix)
        assert self.width == other.width and self.height == other.height
        new = Matrix(self)
        for i in range(len(other)): new[i] += other[i]
        return new

    def __sub__(self, other):
        assert isinstance(other, Matrix)
        assert self.width == other.width and self.height == other.height
        new = Matrix(self)
        for i in range(len(other)): new[i] -= other[i]
        return new

    def __mul__(self, other):
        assert isinstance(other, (Matrix, int, float))
        if not isinstance(other, Matrix):
            new = Matrix(self)
            for row in new:
                row *= other
            return new

        assert self.width == other.height
        # transpose the matrix so its easier to work with
        new = Matrix(width=other.width, height=self.height)
        other = other.transpose

        # loop over each row in the new matrix
        for i in range(new.height):
            # loop over each column in the new matrix
            for j in range(new.width):
                # set the element equal to the dot product of the matrix rows
                new[i][j] = self[i]*other[j]

        return new

    def __truediv__(self, other):
        assert isinstance(other, (Matrix, int, float))
        if not isinstance(other, Matrix):
            new = Matrix(self)
            for row in new:
                row /= other
            return new
        assert self.width == other.height
        return self * other.inverse

    __repr__ = __str__
    __radd__ = __add__
    __rsub__ = __sub__
    __rmul__ = __mul__
    __rtruediv__ = __truediv__

    def __iadd__(self, other):
        assert isinstance(other, Matrix)
        assert self.width == other.width and self.height == other.height
        for i in range(len(other)): self[i] += other[i]
        return self

    def __isub__(self, other):
        assert isinstance(other, Matrix)
        assert self.width == other.width and self.height == other.height
        for i in range(len(other)): self[i] -= other[i]
        return self

    def __imul__(self, other):
        assert isinstance(other, (Matrix, int, float))
        if not isinstance(other, Matrix):
            for row in self:
                row *= other
            return self

        assert self.width == other.height
        # transpose the matrix so its easier to work with
        new = Matrix(width=other.width, height=self.height)
        other = other.transpose

        # loop over each row in the new matrix
        for i in range(new.height):
            # loop over each column in the new matrix
            for j in range(new.width):
                # set the element equal to the dot product of the matrix rows
                new[i][j] = self[i]*other[j]

        # replace the values in this matrix with those in the new matrix
        self.width = 0
        self.height = new.height
        for i in range(self.height):
            self.width = len(self[i])
            self[i] = new[i]

        return self

    def __itruediv__(self, other):
        assert isinstance(other, (Matrix, int, float))
        if not isinstance(other, Matrix):
            for row in self:
                row /= other
            return self
        self *= other.inverse
        return self

    def to_quaternion(self):
        assert self.width == 3
        assert self.height == 3
        return Quaternion(matrix_to_quaternion(self))

    @property
    def determinant(self):
        assert self.width == self.height, "Non-square matrices do not have determinants."
        if self.width == 2:
            return self[0][0] * self[1][1] - self[0][1] * self[1][0]

        d = 0
        sub_matrix = Matrix(width=self.width - 1, height=self.height - 1)
        for i in range(self.width):
            for j in range(sub_matrix.height):
                for k in range(sub_matrix.width):
                    sub_matrix[j][k] = self[j+1][(i + k + 1) % self.width]
            d += self[0][i] * sub_matrix.determinant

        return d

    @property
    def transpose(self):
        transpose = Matrix(width=self.height, height=self.width)
        for r in range(self.height):
            for c in range(self.width):
                transpose[c][r] = self[r][c]
        return transpose

    @property
    def inverse(self, find_best_inverse=True):
        # cannot invert non-square matrices. check for that
        if self.width != self.height:
            raise MatrixNotInvertable("Cannot invert non-square matrix.")
        elif not self.determinant:
            raise MatrixNotInvertable("Matrix is non-invertible.")

        regular, inverse = self.row_reduce(
            Matrix(width=self.width, height=self.height, identity=True),
            find_best_reduction=find_best_inverse
            )

        return inverse

    def row_reduce(self, other, find_best_reduction=True):
        # cant row-reduce if number of columns is greater than number of rows
        assert self.width <= self.height

        # WIDTH NOTE: We will loop over the width rather than height for
        #     several things here, as we do not care about any rows that
        #     don't intersect the columns at a diagonal. Essentially we're
        #     treating a potentially non-square matrix as square(we're
        #     ignoring the higher numbered rows) by rearranging the rows.

        new_row_order = self.get_row_reduce_order(find_best_reduction)
        if new_row_order is None:
            raise CannotRowReduce(
                "Impossible to rearrange rows to row-reduce:\n%s" % self)

        reduced = Matrix(self)
        orig_other = list(other)
        # rearrange rows so diagonals are all non-zero
        for i in range(len(new_row_order)):
            reduced[i] = self[new_row_order[i]]
            other[i] = orig_other[new_row_order[i]]

        orig_reduced = Matrix(reduced)  # TEMP
        for i in range(self.width):  # read note about looping over width
            # divide both matrices by their diagonal values
            div = reduced[i][i]
            if not div:
                raise CannotRowReduce("Impossible to row-reduce.")

            reduced[i] /= div
            other[i] /= div

            # make copies of the rows that we can multiply for subtraction
            reduced_diff = MatrixRow(reduced[i])
            other_diff = MatrixRow(other[i])

            # loop over the rows NOT intersecting the column at the diagonal
            for j in range(self.width):
                if i == j:
                    continue
                # get the value that needs to be subtracted from
                # where this row intersects the current column
                mul = reduced[j][i]

                # subtract the difference row from each of the
                # rows above it to set everything in the column
                # above an below the diagonal intersection to 0
                reduced[j] -= reduced_diff*mul
                other[j] -= other_diff*mul

        return reduced, other

    def get_row_reduce_order(self, find_best_reduction=True):
        nonzero_diag_row_indices = list(set() for i in range(self.width))
        valid_row_orders = {}

        # determine which rows have a nonzero value on each diagonal
        for i in range(self.height):
            for j in range(self.width):
                if self[i][j]:
                    nonzero_diag_row_indices[j].add(i)

        self._get_valid_diagonal_row_orders(
            nonzero_diag_row_indices, valid_row_orders, find_best_reduction)

        # get the highest weighted row order
        test_matrix = Matrix(width=self.width, height=self.width)
        for weight in reversed(sorted(valid_row_orders)):
            for row_order in valid_row_orders[weight]:
                for i in range(len(row_order)):
                    test_matrix[i][:] = self[row_order[i]]

                # make sure the determinant of the matrix made from the
                # row order isn't zero. if it is, the matrix isnt solvable
                if test_matrix.determinant:
                    return row_order

        return None

    def _get_valid_diagonal_row_orders(self, row_indices, row_orders,
                                       choose_best=True, row_order=(),
                                       curr_column=0):
        row_order = list(row_order)
        column_count = len(row_indices)
        if not row_order:
            row_order = [None] * column_count

        # loop over each row with a non-zero value on this diagonal
        for i in row_indices[curr_column]:
            if row_orders and not choose_best:
                # found a valid row arrangement, don't keep checking
                break
            elif i in row_order:
                continue

            row_order[curr_column] = i
            if curr_column + 1 == column_count:
                weight = 1.0
                for j in range(len(row_order)):
                    weight *= abs(self[row_order[j]][j])

                # freeze this row order in place
                row_orders.setdefault(weight, []).append(tuple(row_order))
            else:
                # check the rest of the rows
                self._get_valid_diagonal_row_orders(
                    row_indices, row_orders, choose_best,
                    row_order, curr_column + 1)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from pathlib import Path
from reclaimer.util.matrices import Matrix, matrix_to_quaternion
from reclaimer.model.jms import write_jms, JmsModel, JmsNode, JmsMarker, util

__all__ = ("extract_physics", )


def extract_physics(tagdata, tag_path="", **kw):
    do_write_jms = kw.get('write_jms', True)
    filepath = Path("")
    if do_write_jms:
        filepath = Path(kw.get("out_dir", "")).joinpath(
            Path(tag_path).parent, "physics", "physics.jms")
        if not kw.get('overwrite', True) and filepath.is_file():
            return

    jms_model = JmsModel()
    child_node_ct = 1
    for mp in tagdata.mass_points.STEPTREE:
        child_node_ct = max(child_node_ct, mp.model_node + 1)
        fi, fj, fk = mp.forward
        ui, uj, uk = mp.up
        si, sj, sk = uj*fk - fj*uk, uk*fi - fk*ui, ui*fj - fi*uj

        matrix = Matrix(
            ((fi, fj, fk),
             (si, sj, sk),
             (ui, uj, uk)))
        i, j, k, w = matrix_to_quaternion(matrix)
        # no idea why I have to invert these
        w = -w
        if w < 0:
            i, j, k, w = -i, -j, -k, -w

        jms_model.markers.append(
            JmsMarker(
                mp.name, "physics", -1, mp.model_node, i, j, k, w,
                mp.position.x * 100, mp.position.y * 100, mp.position.z * 100,
                mp.radius * 100,
                ))

    jms_model.nodes = util.generate_fake_nodes(child_node_ct)

    if do_write_jms:
        write_jms(filepath, jms_model)
    else:
        return jms_model



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os

from pathlib import Path
from traceback import format_exc

from reclaimer.sounds.blam_sound_permutation import BlamSoundPermutation
from reclaimer.sounds import constants, ogg, adpcm


class BlamSoundPitchRange:
    name = ""
    _permutations = ()

    def __init__(self):
        self._permutations = {}

    @property
    def permutations(self):
        return self._permutations

    def generate_mouth_data(self):
        for perm in self.permutations.values():
            perm.generate_mouth_data()

    def compress_samples(
            self, compression=constants.COMPRESSION_PCM_16_LE,
            sample_rate=None, encoding=None, chunk_size=None,
            **compressor_kwargs):
        for perm in self.permutations.values():
            perm.compress_samples(compression, sample_rate, encoding,
                                  chunk_size, **compressor_kwargs)

    def regenerate_source(self):
        for perm in self.permutations.values():
            perm.regenerate_source()

    @staticmethod
    def create_from_directory(directory):
        try:
            new_pitch_range = BlamSoundPitchRange()
            new_pitch_range.import_from_directory(directory)
        except Exception:
            print(format_exc())
            new_pitch_range = None

        return new_pitch_range

    def export_to_directory(self, directory, overwrite=False,
                            export_source=True, decompress=True):
        for name, perm in self.permutations.items():
            perm.export_to_file(
                Path(directory, name), overwrite,
                export_source, decompress)

    def import_from_directory(self, directory, clear_existing=True,
                              replace_existing=True):
        if clear_existing:
            self.permutations.clear()

        for root, _, files in os.walk(str(directory)):
            # import each sound file to a new sound permutation
            for filename in files:
                filepath = Path(root, filename)
                if filepath.suffix.lower() != ".wav":
                    # only import wav files
                    continue

                name_key = filepath.stem.lower().strip()
                perm = BlamSoundPermutation.create_from_file(filepath)
                perm.name = filepath.stem.strip()

                if (perm and perm.source_sample_data) and (
                    replace_existing or name_key not in self.permutations):
                    self.permutations[name_key] = perm

            # only loop over the files
            break


class BlamSoundBank:
    # processing settings
    encoding = constants.ENCODING_MONO
    compression = constants.COMPRESSION_PCM_16_LE
    sample_rate = constants.SAMPLE_RATE_22K
    split_into_smaller_chunks = True

    # chunk_size is capped to constants.MAX_SAMPLE_CHUNK_SIZE
    # this value is fine to bump under most circumstances.
    chunk_size = constants.DEF_SAMPLE_CHUNK_SIZE

    vorbis_bitrate_info = None

    adpcm_noise_shaping = adpcm.NOISE_SHAPING_OFF
    adpcm_lookahead = 3

    _pitch_ranges = ()

    def __init__(self):
        self._pitch_ranges = {}
        self.vorbis_bitrate_info = ogg.VorbisBitrateInfo()

    @property
    def pitch_ranges(self):
        return self._pitch_ranges

    def generate_mouth_data(self):
        for pitch_range in self.pitch_ranges.values():
            pitch_range.generate_mouth_data()

    def compress_samples(self):
        chunk_size = None
        if self.split_into_smaller_chunks:
            chunk_size = self.chunk_size

        adpcm_kwargs = dict(
            noise_shaping=self.adpcm_noise_shaping,
            lookahead=self.adpcm_lookahead
            )
        ogg_kwargs = dict(
            bitrate_info=self.vorbis_bitrate_info
            )

        for pitch_range in self.pitch_ranges.values():
            pitch_range.compress_samples(
                self.compression, self.sample_rate, self.encoding,
                chunk_size, adpcm_kwargs=adpcm_kwargs, ogg_kwargs=ogg_kwargs,
                )

    def regenerate_source(self):
        for pitch_range in self.pitch_ranges.values():
            pitch_range.regenerate_source()

    @staticmethod
    def create_from_directory(directory):
        try:
            new_sound_bank = BlamSoundBank()
            new_sound_bank.import_from_directory(directory)
        except Exception:
            print(format_exc())
            new_sound_bank = None

        return new_sound_bank

    def export_to_directory(self, directory, overwrite=False,
                            export_source=True, decompress=True):
        for name, pitch_range in self.pitch_ranges.items():
            if len(self.pitch_ranges) > 1:
                pitch_directory = Path(directory, name)
            else:
                pitch_directory = Path(directory)

            pitch_range.export_to_directory(
                pitch_directory, overwrite, export_source, decompress)

    def import_from_directory(self, directory, clear_existing=True,
                              merge_same_names=False):
        if clear_existing:
            self.pitch_ranges.clear()

        # try to load default pitch range in sound root
        default_pitch_range = BlamSoundPitchRange.create_from_directory(
            directory)

        # try to load default from a directory named default as well
        default_dir_pitch_range = BlamSoundPitchRange.create_from_directory(
            Path(directory, constants.DEFAULT_PITCH_RANGE_NAME))

        # either update the root default pitch ranges with the
        # other, or replace it with the directory one
        if default_dir_pitch_range and default_dir_pitch_range.permutations:
            if default_pitch_range and default_pitch_range.permutations:
                default_pitch_range.permutations.update(
                    default_dir_pitch_range.permutations)
            else:
                default_pitch_range = default_dir_pitch_range

        # merge the default pitch range into this sound bank
        if default_pitch_range:
            default_pitch_range.name = constants.DEFAULT_PITCH_RANGE_NAME
            name_key = constants.DEFAULT_PITCH_RANGE_NAME

            if not default_pitch_range.permutations:
                pass
            elif self.pitch_ranges.get(name_key) is None:
                self.pitch_ranges[name_key] = default_pitch_range
            elif merge_same_names:
                self.pitch_ranges[name_key].permutations.update(
                    default_pitch_range.permutations)

        for root, dirs, _ in os.walk(directory):
            for dirname in dirs:
                name_key = dirname.lower().strip()
                if name_key == constants.DEFAULT_PITCH_RANGE_NAME:
                    # skip default pitch range(we took care of it above)
                    continue

                pitch_range = BlamSoundPitchRange.create_from_directory(
                    Path(root, dirname))
                pitch_range.name = dirname.strip()

                # merge the pitch range into this sound bank
                if pitch_range and pitch_range.permutations:
                    if self.pitch_ranges.get(name_key) is None:
                        self.pitch_ranges[name_key] = pitch_range
                    elif merge_same_names:
                        self.pitch_ranges[name_key].permutations.update(
                            pitch_range.permutations)

            # only looping over the first level directories
            break



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: -DeToX-
# 	Created layout of plugin
# revision: 2		author: DeadCanadian
# 	naming tags
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


bsdt_unknown0 = Struct("unknown0",
    BytesRaw("unknown_0", SIZE=8, VISIBLE=False),
    h3_rawdata_ref("unknown_1", VISIBLE=False),
    BytesRaw("unknown_2", SIZE=8, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=36
    )


bsdt_unknown1 = Struct("unknown1",
    BytesRaw("unknown_0", SIZE=8, VISIBLE=False),
    h3_rawdata_ref("unknown_1", VISIBLE=False),
    BytesRaw("unknown_2", SIZE=8, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=36
    )


bsdt_body = Struct("tagdata",
    Float("maximum_vitality"),
    h3_dependency("effect"),
    h3_dependency("sound"),
    BytesRaw("unknown_0", SIZE=16, VISIBLE=False),
    h3_dependency("crack_bitmap"),
    h3_dependency("hole_bitmap"),
    BytesRaw("unknown_1", SIZE=36, VISIBLE=False),
    h3_reflexive("unknown0", bsdt_unknown0),
    BytesRaw("unknown_2", SIZE=12, VISIBLE=False),
    h3_reflexive("unknown1", bsdt_unknown1),
    BytesRaw("unknown_3", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=160
    )


def get():
    return bsdt_def

bsdt_def = TagDef("bsdt",
    h3_blam_header('bsdt'),
    bsdt_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["bsdt"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.constants import *
from reclaimer.util import fourcc_to_int


# maps tag class four character codes(fccs) in
# their string encoding to their int encoding.
h3_tag_class_fcc_to_be_int = {}
h3_tag_class_fcc_to_le_int = {}
# maps tag class four character codes(fccs) in
# their int encoding to their string encoding.
h3_tag_class_be_int_to_fcc = {}
h3_tag_class_le_int_to_fcc = {}


HALO3_SHARED_MAP_TYPES = ("mainmenu", "shared", "campaign")

# maps tag class four character codes to the tags file extension
# 177 tag classes
nul_char = b"\x00".decode(encoding="latin-1")
h3_tag_class_fcc_to_ext = {
    nul_char + "rmc": "shader_contrail",
    nul_char + "rmp": "shader_particle",
    "$#!+": "cache_file_sound",
    "*cen": "scenario_scenery_resource",
    "*eap": "scenario_weapons_resource",
    "*ehi": "scenario_vehicles_resource",
    "*fsc": "scenario_effect_scenery_resource",
    "*igh": "scenario_lights_resource",
    "*ipd": "scenario_bipeds_resource",
    "*qip": "scenario_equipment_resource",
    "*rea": "scenario_creature_resource",
    "*sce": "scenario_sound_scenery_resource",
    "/**/": "scenario_comments_resource",
    "<fx>": "sound_effect_template",
    "BooM": "stereo_system",
    "adlg": "ai_dialogue_globals",
    "ai**": "scenario_ai_resource",
    "ant!": "antenna",
    "beam": "beam_system",
    "bink": "bink",
    "bipd": "biped",
    "bitm": "bitmap",
    "bkey": "gui_button_key_definition",
    "bloc": "crate",
    "bmp3": "gui_bitmap_widget_definition",
    "bsdt": "breakable_surface",
    "cddf": "collision_damage",
    "cfxs": "camera_fx_settings",
    "chad": "chud_animation_definition",
    "char": "character",
    "chdt": "chud_definition",
    "chgd": "chud_globals_definition",
    "chmt": "chocolate_mountain_new",
    "cin*": "scenario_cinematics_resource",
    "cine": "cinematic",
    "cisc": "cinematic_scene",
    "clu*": "scenario_cluster_data_resource",
    "clwd": "cloth",
    "cntl": "contrail_system",
    "coll": "collision_model",
    "colo": "color_table",
    "crea": "creature",
    "crte": "cortana_effect_definition",
    "ctrl": "device_control",
    "cub*": "scenario_cubemap_resource",
    "dc*s": "scenario_decorators_resource",
    "dctr": "decorator_set",
    "dec*": "scenario_decals_resource",
    "decs": "decal_system",
    "devi": "device",
    "devo": "cellular_automata",
    "dgr*": "scenario_devices_resource",
    "dobc": "detail_object_collection",
    "draw": "rasterizer_cache_file_globals",
    "drdf": "damage_response_definition",
    "dsrc": "gui_datasource_definition",
    "effe": "effect",
    "effg": "effect_globals",
    "efsc": "effect_scenery",
    "egor": "screen_effect",
    "eqip": "equipment",
    "flck": "flock",
    "fldy": "fluid_dynamics",
    "fog ": "planar_fog",
    "foot": "material_effects",
    "fpch": "patchy_fog",
    "frag": "fragment",
    "gint": "giant",
    "glps": "global_pixel_shader",
    "glvs": "global_vertex_shader",
    "goof": "multiplayer_variant_settings_interface_definition",
    "grup": "gui_group_widget_definition",
    "hlmt": "model",
    "hlsl": "hlsl_include",
    "hsc*": "scenario_hs_source_file",
    "item": "item",
    "itmc": "item_collection",
    "jmad": "model_animation_graph",
    "jmrq": "sandbox_text_value_pair_definition",
    "jpt!": "damage_effect",
    "lens": "lens_flare",
    "ligh": "light",
    "lsnd": "sound_looping",
    "lst3": "gui_list_widget_definition",
    "lswd": "leaf_system",
    "ltvl": "light_volume_system",
    "mach": "device_machine",
    "matg": "globals",
    "mdl3": "gui_model_widget_definition",
    "mdlg": "ai_mission_dialogue",
    "metr": "meter",
    "mffn": "muffin",
    "mode": "render_model",
    "mply": "multiplayer_scenario_description",
    "mulg": "multiplayer_globals",
    "nclt": "new_cinematic_lighting",
    "obje": "object",
    "perf": "performance_throttles",
    "phmo": "physics_model",
    "pixl": "pixel_shader",
    "play": "cache_file_resource_layout_table",
    "pmdf": "particle_model",
    "pmov": "particle_physics",
    "pphy": "point_physics",
    "proj": "projectile",
    "prt3": "particle",
    "rasg": "rasterizer_globals",
    "rm  ": "render_method",
    "rmb ": "shader_beam",
    "rmcs": "shader_custom",
    "rmct": "shader_cortana",
    "rmd ": "shader_decal",
    "rmdf": "render_method_definition",
    "rmfl": "shader_foliage",
    "rmhg": "shader_halogram",
    "rmlv": "shader_light_volume",
    "rmop": "render_method_option",
    "rmsh": "shader",
    "rmsk": "shader_skin",
    "rmt2": "render_method_template",
    "rmtr": "shader_terrain",
    "rmw ": "shader_water",
    "rwrd": "render_water_ripple",
    "sFdT": "scenario_faux_data",
    "sLdT": "scenario_lightmap",
    "sbsp": "scenario_structure_bsp",
    "scen": "scenery",
    "scn3": "gui_screen_widget_definition",
    "scnr": "scenario",
    "sddt": "structure_design",
    "sefc": "area_screen_effect",
    "sfx+": "sound_effect_collection",
    "sgp!": "sound_global_propagation",
    "shit": "shield_impact",
    "sily": "text_value_pair_definition",
    "skn3": "gui_skin_definition",
    "sky*": "scenario_sky_references_resource",
    "skya": "sky_atm_parameters",
    "smap": "shared_cache_file_layout",
    "sncl": "sound_classes",
    "snd!": "sound",
    "snde": "sound_environment",
    "snmx": "sound_mix",
    "spk!": "sound_dialogue_constants",
    "ssce": "sound_scenery",
    "sslt": "scenario_structure_lighting_resource",
    "stli": "scenario_structure_lighting_info",
    "stse": "structure_seams",
    "styl": "style",
    "term": "device_terminal",
    "trak": "camera_track",
    "trg*": "scenario_trigger_volumes_resource",
    "txt3": "gui_text_widget_definition",
    "udlg": "dialogue",
    "ugh!": "sound_cache_file_gestalt",
    "uise": "user_interface_sounds_definition",
    "unic": "multilingual_unicode_string_list",
    "unit": "unit",
    "vehc": "vehicle_collection",
    "vehi": "vehicle",
    "vtsh": "vertex_shader",
    "wacd": "gui_widget_animation_collection_definition",
    "wclr": "gui_widget_color_animation_definition",
    "weap": "weapon",
    "wezr": "game_engine_settings_definition",
    "wfon": "gui_widget_font_animation_definition",
    "wgan": "gui_widget_animation_definition",
    "wgtz": "user_interface_globals_definition",
    "whip": "cellular_automata2d",
    "wigl": "user_interface_shared_globals_definition",
    "wind": "wind",
    "wpos": "gui_widget_position_animation_definition",
    "wrot": "gui_widget_rotation_animation_definition",
    "wscl": "gui_widget_scale_animation_definition",
    "wspr": "gui_widget_sprite_animation_definition",
    "wtuv": "gui_widget_texture_coordinate_animation_definition",
    "zone": "cache_file_resource_gestalt",
    }

for tag_cls in h3_tag_class_fcc_to_ext:
    h3_tag_class_fcc_to_be_int[tag_cls] = fourcc_to_int(tag_cls)
    h3_tag_class_be_int_to_fcc[fourcc_to_int(tag_cls)] = tag_cls
    h3_tag_class_fcc_to_le_int[tag_cls] = fourcc_to_int(tag_cls)
    h3_tag_class_le_int_to_fcc[fourcc_to_int(tag_cls)] = tag_cls



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from math import cos

from reclaimer.hek.defs.objs.tag import HekTag

class LensTag(HekTag):

    def calc_internal_data(self):
        HekTag.calc_internal_data(self)

        tagdata = self.data.tagdata
        tagdata.cosine_falloff_angle = cos(tagdata.falloff_angle)
        tagdata.cosine_cutoff_angle  = cos(tagdata.cutoff_angle)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import math
try:
    #import collada
    # Null the import to collada since we're not using it
    collada = None
except ImportError:
    collada = None

from reclaimer.model.jms import JmsNode, JmsMaterial, JmsMarker,\
     JmsVertex, JmsTriangle, JmsModel
from traceback import format_exc


special_mat_names = (
    "sky", "seamsealer", "portal", "exactportal",
    "weatherpoly", "sound", "unused")

def jms_model_from_dae(filepath, model_name=None):
    if collada is None:
        return

    if model_name is None:
        model_name = "__unnamed"

    collada_model = collada.Collada(filepath)
    print(collada_model.geometries)

    model = JmsModel(model_name)

    regions = model.regions
    mats = model.materials
    nodes = model.nodes
    tris = model.tris
    verts = model.verts

    for collada_mat in collada_model.materials:
        mat_name = collada_mat.name

        # correct DAE not being able to use + in material names
        if mat_name.startswith('_'):
            mat_name = mat_name.lstrip("_")
            if mat_name in special_mat_names:
                mat_name = "+" + mat_name

        mats.append(JmsMaterial(mat_name))

    if not mats:
        mats.append(JmsMaterial("__unnamed"))

    return
    # idk if i'll ever implement this. collada is just too complex

    return model



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from ...hek.defs.objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

efpp_attrs = Struct("efpp_attrs",
    Pad(12),
    QStruct("quad_tesselation",
        SInt16("x"), SInt16("y"), ORIENT='h'
        ),
    from_to_zero_to_one("x_screen_bounds"),
    from_to_zero_to_one("y_screen_bounds"),
    SIZE=60
    )


efpp_body = Struct("tagdata", efpp_attrs)

def get():
    return efpp_def

efpp_def = TagDef("efpp",
    blam_header_os('efpp'),
    efpp_body,

    ext=".effect_postprocess", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef


frequency_vibration = Struct("",
    float_zero_to_one("frequency"),
    float_sec("duration"),
    SEnum16("fade_function", *fade_functions),
    )

jpt__body = Struct("tagdata",
    from_to_wu("radius"),
    float_zero_to_one("cutoff_scale"),
    Bool32("flags",
        "dont_scale_by_distance",
        ),
    Pad(20),

    #Screen_Flash
    Struct("screen_flash",
        SEnum16("type",
            "none",
            "lighten",
            "darken",
            "max",
            "min",
            "invert",
            "tint",
            ),
        SEnum16("priority",
            "low",
            "medium",
            "high",
            ),
        Pad(12),

        float_sec("duration"),
        SEnum16("fade_function", *fade_functions),
        Pad(10),

        float_zero_to_one("maximum_intensity"),
        Pad(4),

        QStruct("tint_lower_bound", INCLUDE=argb_float),
        ),

    Struct("low_frequency_vibrate", INCLUDE=frequency_vibration),
    Pad(10),
    Struct("high_frequency_vibrate", INCLUDE=frequency_vibration),
    Pad(30),

    Struct("temporary_camera_impulse",
        float_sec("duration"),
        SEnum16("fade_function", *fade_functions),
        Pad(2),

        float_rad("rotation"),  # radians
        float_wu("pushback"),
        from_to_wu("jitter"),
        Pad(8),
        ),

    float_rad("permanent_camera_impulse_angle"),
    Pad(16),

    Struct("camera_shaking",
        float_sec("duration"),
        SEnum16("fade_function", *fade_functions),
        Pad(2),

        float_wu("random_translation"),
        float_rad("random_rotation"),  # radians
        Pad(12),

        SEnum16("wobble_function", *animation_functions),
        Pad(2),
        float_sec("wobble_function_period"),
        Float("wobble_weight"),
        Pad(32),
        ),

    dependency("sound", "snd!"),
    Pad(112),

    QStruct("breaking_effect",
        float_wu_sec("forward_velocity"),
        float_wu("forward_radius"),
        Float("forward_exponent"),
        Pad(12),

        float_wu_sec("outward_velocity"),
        float_wu("outward_radius"),
        Float("outward_exponent"),
        Pad(12),
        ),

    Struct("damage",
        SEnum16("priority",
            "none",
            "harmless",
            {NAME: "backstab", GUI_NAME: "lethal to the unsuspecting"},
            "emp",
            ),
        SEnum16("category", *damage_category),
        Bool32("flags",
            "does_not_hurt_owner",
            {NAME: "headshot", GUI_NAME: "causes headshots"},
            "pings_resistant_units",
            "does_not_hurt_friends",
            "does_not_ping_units",
            "detonates_explosives",
            "only_hurts_shields",
            "causes_flaming_death",
            {NAME: "indicator_points_down",
             GUI_NAME: "damage indicator always points down"},
            "skips_shields",
            "only_hurts_one_infection_form",
            {NAME: "multiplayer_headshot",
             GUI_NAME: "causes multiplayer headshots"},
            "infection_form_pop",
            ),
        float_wu("aoe_core_radius"),
        Float("damage_lower_bound"),
        QStruct("damage_upper_bound", INCLUDE=from_to),
        float_zero_to_one("vehicle_passthrough_penalty"),
        float_zero_to_one("active_camouflage_damage"),
        float_zero_to_one("stun"),
        float_zero_to_one("maximum_stun"),
        float_sec("stun_time"),
        Pad(4),
        float_zero_to_inf("instantaneous_acceleration"),
        Pad(8),
        ),

    damage_modifiers,
    SIZE=672,
    )


def get():
    return jpt__def

jpt__def = TagDef("jpt!",
    blam_header('jpt!', 6),
    jpt__body,

    ext=".damage_effect", endian=">", tag_cls=HekTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os
import zlib

from pathlib import Path
from traceback import format_exc

from reclaimer.meta.wrappers.halo_map import HaloMap
from reclaimer import data_extraction
from reclaimer.h2.util import HALO2_MAP_TYPES, split_raw_pointer
from reclaimer.util import is_reserved_tag, int_to_fourcc
from reclaimer.meta.wrappers.string_id_manager import StringIdManager
from reclaimer.meta.wrappers.tag_index_manager import TagIndexManager
from reclaimer.meta.wrappers.tag_index_converters import h2_alpha_to_h1_tag_index,\
     h2_to_h1_tag_index, h3_to_h1_tag_index

from supyr_struct.util import is_path_empty

from supyr_struct.buffer import BytearrayBuffer


class Halo2Map(HaloMap):
    '''Generation 2 map'''
    ugh__meta = None

    tag_defs_module = "reclaimer.h2.defs"
    tag_classes_to_load = (
        "ant!", "bitm", "hsc*", "pphy", "snd!", "shad", "trak", "ugh!"
        )
    data_extractors = data_extraction.h2_data_extractors

    def __init__(self, maps=None):
        HaloMap.__init__(self, maps)
        self.setup_tag_headers()

    def inject_rawdata(self, meta, tag_cls, tag_index_ref):
        # get some rawdata that would be pretty annoying to do in the parser
        if tag_cls == "bitm":
            # grab bitmap data correctly from map
            new_pixels = BytearrayBuffer()
            pix_off = 0

            for bitmap in meta.bitmaps.STEPTREE:
                # grab the bitmap data from the correct map
                bitmap.pixels_offset = pix_off

                ptr, map_name = split_raw_pointer(bitmap.lod1_offset)
                halo_map = self
                if map_name != "local":
                    halo_map = self.maps.get(map_name)

                if halo_map is None:
                    bitmap.lod1_size = 0
                    continue

                halo_map.map_data.seek(ptr)
                mip_pixels = halo_map.map_data.read(bitmap.lod1_size)
                mip_pixels = zlib.decompress(mip_pixels)

                new_pixels += mip_pixels
                bitmap.lod1_size = len(mip_pixels)
                pix_off += bitmap.lod1_size

            meta.processed_pixel_data.STEPTREE = new_pixels

    def meta_to_tag_data(self, meta, tag_cls, tag_index_ref, **kwargs):
        engine     = self.engine
        tag_index  = self.tag_index

        if tag_cls == "bitm":
            # set the size of the compressed plate data to nothing
            meta.compressed_color_plate_data.STEPTREE = BytearrayBuffer()
            new_pixels_offset = 0

            # uncheck the prefer_low_detail flag and
            # set up the lod1_offset correctly.
            for bitmap in meta.bitmaps.STEPTREE:
                bitmap.lod1_offset = new_pixels_offset
                new_pixels_offset += bitmap.lod1_size

                bitmap.lod2_offset = bitmap.lod3_offset = bitmap.lod4_offset =\
                                     bitmap.lod5_offset = bitmap.lod6_offset = 0
                bitmap.lod2_size = bitmap.lod3_size = bitmap.lod4_size =\
                                   bitmap.lod5_size = bitmap.lod6_size = 0

        return meta

    def get_resource_map_paths(self, maps_dir=""):
        map_paths = {name: None for name in HALO2_MAP_TYPES[1:]}
        if self.engine == "halo2alpha":
            map_paths.pop("single_player_shared", None)

        if not is_path_empty(maps_dir):
            maps_dir = Path(maps_dir)
        else:
            maps_dir = self.filepath.parent

        # detect the map paths for the resource maps
        for map_name in sorted(map_paths):
            map_path = str(maps_dir.joinpath(map_name))
            if self.maps.get(map_name) is not None:
                map_path = self.maps[map_name].filepath
            elif os.path.exists(map_path + ".map"):
                map_path = Path(map_path + ".map")
            elif os.path.exists(map_path + "_DECOMP.map"):
                map_path = Path(map_path + "_DECOMP.map")
            elif os.path.exists(map_path + ".map.dtz"):
                map_path = Path(map_path + ".map.dtz")
            else:
                map_path = None

            map_paths[map_name] = map_path

        return map_paths

    def load_map(self, map_path, **kwargs):
        HaloMap.load_map(self, map_path, **kwargs)
        tag_index = self.tag_index
        if self.engine == "halo2alpha":
            self.tag_index = h2_alpha_to_h1_tag_index(self.map_header, tag_index)
        else:
            self.tag_index = h2_to_h1_tag_index(self.map_header, tag_index)

        # add the tag data section
        self.map_pointer_converter.add_page_info(
            self.index_magic, self.map_header.tag_index_header_offset,
            self.map_header.tag_index_data_size + self.map_header.tag_data_size
            )

        self.string_id_manager = StringIdManager(
            self.map_header.strings.string_id_table, (),
            )
        self.tag_index_manager = TagIndexManager(self.tag_index.tag_index)
        map_type = self.map_header.map_type.data - 1
        if map_type > 0 and map_type < 4:
            self.is_resource = True
            self.maps[HALO2_MAP_TYPES[map_type]] = self

        # get the sound_cache_file_gestalt meta
        try:
            if self.engine == "halo2vista":
                ugh__id = None
                for b in self.tag_index.tag_index:
                    if int_to_fourcc(b.class_1.data) == "ugh!":
                        ugh__id = b.id & 0xFFff
                        break

                self.ugh__meta = self.get_meta(ugh__id)
                if self.ugh__meta is None:
                    print("Could not read sound_cache_file_gestalt tag")
        except Exception:
            print(format_exc())
            print("Could not read sound_cache_file_gestalt tag")

        self.clear_map_cache()

    def get_meta(self, tag_id, reextract=False, **kw):
        if tag_id is None:
            return
        elif self.engine == "halo2alpha":
            return

        scnr_id = self.orig_tag_index.scenario_tag_id & 0xFFff
        tag_index_array = self.tag_index.tag_index
        shared_map    = self.maps.get("shared")

        tag_id &= 0xFFFF
        tag_index_ref = self.tag_index_manager.get_tag_index_ref(tag_id)
        if tag_index_ref is None:
            return

        matg_id = self.orig_tag_index.globals_tag_id & 0xFFff
        sp_shared_map = self.maps.get("single_player_shared")

        if tag_id >= 10000 and shared_map is not self:
            if shared_map is None: return
            return shared_map.get_meta(tag_id, reextract)
        elif tag_id >= len(tag_index_array) and sp_shared_map is not self:
            if sp_shared_map is None: return
            return sp_shared_map.get_meta(tag_id, reextract)

        tag_index_ref = tag_index_array[tag_id]

        tag_cls = None
        if   tag_id == scnr_id: tag_cls = "scnr"
        elif tag_id == matg_id: tag_cls = "matg"
        elif tag_index_ref.class_1.enum_name not in ("<INVALID>", "NONE"):
            tag_cls = int_to_fourcc(tag_index_ref.class_1.data)

        desc = self.get_meta_descriptor(tag_cls)
        if desc is None or tag_cls is None:        return
        elif reextract:                            pass
        elif tag_id == scnr_id and self.scnr_meta: return self.scnr_meta
        elif tag_id == matg_id and self.matg_meta: return self.matg_meta
        elif tag_cls == "ugh!" and self.ugh__meta: return self.ugh__meta

        block = [None]
        offset = self.map_pointer_converter.v_ptr_to_f_ptr(
            tag_index_ref.meta_offset)

        try:
            # read the meta data from the map
            desc['TYPE'].parser(
                desc, parent=block, attr_index=0, rawdata=self.map_data,
                map_string_id_manager=self.string_id_manager,
                map_pointer_converter=self.map_pointer_converter,
                tag_index_manager=self.tag_index_manager, offset=offset,)
        except Exception:
            print(format_exc())
            if kw.get("allow_corrupt"):
                return block[0]
            return

        meta = block[0]
        try:
            self.record_map_cache_read(tag_id, 0)
            if self.map_cache_over_limit():
                self.clear_map_cache()

            self.inject_rawdata(meta, tag_cls, tag_index_ref)
            if tag_cls == "ugh!":
                self.ugh__meta = meta
        except Exception:
            print(format_exc())
            if not kw.get("allow_corrupt"):
                meta = None

        return meta

    def generate_map_info_string(self):
        string = HaloMap.generate_map_info_string(self)
        string += """

Calculated information:
    index magic == %s
    map magic   == %s
""" % (self.index_magic, self.map_magic)

        if self.engine == "halo2alpha":
            string += (("""
Tag index:
    tag count           == %s
    scenario tag id     == %s
    index array pointer == %s""") %
            (self.orig_tag_index.tag_count,
             self.orig_tag_index.scenario_tag_id & 0xFFff,
             self.tag_index.tag_index_offset))
        else:
            used_tag_count = 0
            local_tag_count = 0
            for index_ref in self.tag_index.tag_index:
                if is_reserved_tag(index_ref):
                    continue
                elif index_ref.meta_offset != 0:
                    local_tag_count += 1
                used_tag_count += 1

            string += (("""
Tag index:
    tag count           == %s
    used tag count      == %s
    local tag count     == %s
    tag types count     == %s
    scenario tag id     == %s
    globals  tag id     == %s
    index array pointer == %s""") %
            (self.orig_tag_index.tag_count, used_tag_count, local_tag_count,
             self.orig_tag_index.tag_types_count,
             self.orig_tag_index.scenario_tag_id & 0xFFff,
             self.orig_tag_index.globals_tag_id & 0xFFff,
             self.tag_index.tag_index_offset))

        return string



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Copypasta H2.
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

sncl_sound_classe_cache_miss_mode = (
    "discard",
    "postpone",
    )

sncl_sound_classe_stereo_playback_type = (
    "first_person",
    "ambient",
    )


sncl_sound_classe = Struct("sound_classe",
    SInt16("max_sounds_per_tag"),
    SInt16("max_sounds_per_object"),
    SInt32("preemption_time"),
    Bool16("internal_flags",
        "valid",
        "is_speech",
        "scripted",
        "stops_with_object",
        ("multilingual", 1 << 7),
        ),
    Bool16("flags",
        "plays_during_pause",
        "dry_stereo_mix",
        "no_object_obstruction",
        "use_center_speaker_unspatialized",
        "send_mono_to_lfe",
        "deterministic",
        "use_huge_transmission",
        "always_use_speakers",
        "dont_strip_from_mainmenu",
        "ignore_stereo_headroom",
        ),
    SInt16("priority"),
    SEnum16("cache_miss_mode", *sncl_sound_classe_cache_miss_mode),
    Float("reverb_gain"),
    BytesRaw("unknown_0", SIZE=12, VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    BytesRaw("unknown_3", SIZE=8, VISIBLE=False),
    Float("unknown_4", VISIBLE=False),
    Float("unknown_5", VISIBLE=False),
    Float("unknown_6", VISIBLE=False),
    QStruct("distance_bounds", INCLUDE=from_to),
    QStruct("gain_bounds", INCLUDE=from_to),
    Float("cutscene_ducking"),
    Float("cutscene_ducking_fade_in_time"),
    Float("cutscene_ducking_sustain"),
    Float("cutscene_ducking_fade_out_time"),
    Float("scripted_dialog_ducking"),
    Float("scripted_dialog_ducking_fade_in"),
    BytesRaw("unknown_7", SIZE=4, VISIBLE=False),
    Float("unknown_8", VISIBLE=False),
    Float("unknown_9", VISIBLE=False),
    Float("unknown_10", VISIBLE=False),
    BytesRaw("unknown_11", SIZE=4, VISIBLE=False),
    Float("unknown_12", VISIBLE=False),
    Float("unknown_13", VISIBLE=False),
    Float("doppler_factor"),
    SEnum8("stereo_playback_type", *sncl_sound_classe_stereo_playback_type),
    SInt8("unknown_14", VISIBLE=False),
    SInt8("unknown_15", VISIBLE=False),
    SInt8("unknown_16", VISIBLE=False),
    Float("transmission_multiplier"),
    Float("obstruction_max_bend"),
    Float("occlusion_max_bend"),
    SInt32("unknown_17", VISIBLE=False),
    ENDIAN=">", SIZE=152
    )


sncl_body = Struct("tagdata",
    h3_reflexive("sound_classes", sncl_sound_classe),
    ENDIAN=">", SIZE=12
    )


def get():
    return sncl_def

sncl_def = TagDef("sncl",
    h3_blam_header('sncl'),
    sncl_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["sncl"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.plac import *

#import and use the open saucified obje attrs
from .obje import *

# replace the object_type enum one that uses
# the correct default value for this object
obje_attrs = dict(obje_attrs)
obje_attrs[0] = dict(obje_attrs[0], DEFAULT=10)

plac_body = dict(plac_body)
plac_body[0] = obje_attrs

def get():
    return plac_def

plac_def = TagDef("plac",
    blam_header_stubbs('plac', 2),
    plac_body,

    ext=".placeholder", endian=">", tag_cls=ObjeTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

'''
Contains functions for quickly reading/writing certain structs in mapfiles.
This is very unstructured as it needs to be as fast as possible, so
raw offsets and map magic must be provided explicitely.
'''

from types import MethodType
from struct import pack, unpack

from reclaimer.constants import \
     tag_class_be_int_to_fcc_os,     tag_class_fcc_to_ext_os,\
     tag_class_be_int_to_fcc_stubbs, tag_class_fcc_to_ext_stubbs

tag_cls_int_to_fcc = dict(tag_class_be_int_to_fcc_os)
tag_cls_int_to_ext = {}

tag_cls_int_to_fcc.update(tag_class_be_int_to_fcc_stubbs)

for key in tag_class_be_int_to_fcc_os:
    tag_cls_int_to_ext[key] = tag_class_fcc_to_ext_os[
        tag_class_be_int_to_fcc_os[key]]

for key in tag_class_be_int_to_fcc_stubbs:
    tag_cls_int_to_ext[key] = tag_class_fcc_to_ext_stubbs[
        tag_class_be_int_to_fcc_stubbs[key]]


NULL_CLASS = b'\xFF\xFF\xFF\xFF'

shader_class_bytes = (
    NULL_CLASS,
    NULL_CLASS,
    NULL_CLASS,
    b'vnes',
    b'osos',
    b'rtos',
    b'ihcs',
    b'xecs',
    b'taws',
    b'algs',
    b'tems',
    b'alps',
    b'rdhs'
    )

object_class_bytes = (
    b'dpib',
    b'ihev',
    b'paew',
    b'piqe',
    b'brag',
    b'jorp',
    b'necs',
    b'hcam',
    b'lrtc',
    b'ifil',
    b'calp',
    b'ecss',
    b'ejbo'
    )

pack_4_uint32 = MethodType(pack, "<LLLL")
unpack_3_uint32 = MethodType(unpack, "<LLL")
unpack_4_uint32 = MethodType(unpack, "<LLLL")
unpack_5_uint32 = MethodType(unpack, "<LLLLL")


def read_reflexive(map_data, refl_offset, max_count=0xFFffFFff,
                   struct_size=1, tag_magic=None, unpacker=unpack_3_uint32):
    '''
    Reads a reflexive from the given map_data at the given offset.
    Returns the reflexive's offset and pointer.
    '''
    map_data.seek(refl_offset)
    count, start, id = unpacker(map_data.read(12))
    if tag_magic is not None:
        map_data.seek(0, 2)
        max_count = min(max_count,
                        (map_data.tell() - (start - tag_magic)) // struct_size)
    return min(count, max_count), start, id


def read_rawdata_ref(map_data, ref_offset, tag_magic=None,
                     unpacker=unpack_5_uint32):
    map_data.seek(ref_offset)
    size, flags, raw_pointer, pointer, id = unpacker(map_data.read(20))
    if tag_magic is not None:
        map_data.seek(0, 2)
        size = min(size, map_data.tell() - (pointer - tag_magic))
    return size, flags, raw_pointer, pointer, id


def move_rawdata_ref(map_data, raw_ref_offset, magic, engine, diffs_by_offsets,
                     unpacker=unpack_4_uint32, packer=pack_4_uint32):
    map_data.seek(raw_ref_offset - magic)
    size, flags, raw_ptr, ptr = unpacker(map_data.read(16))
    if not size or ((flags & 1) and "xbox" not in engine):
        return

    ptr_diff = 0
    # find the highest pointer that is still at or below this rawdatas pointer
    for off in diffs_by_offsets:
        if off <= raw_ptr:
            ptr_diff = max(diffs_by_offsets[off], ptr_diff)

    if not ptr_diff:
        return

    # TODO: Determine if we need to modify the regular pointer
    # and also determine in what way to modify it

    map_data.seek(0, 2)
    raw_ptr += ptr_diff
    assert raw_ptr + size <= map_data.tell()
    map_data.seek(raw_ref_offset - magic)
    map_data.write(packer(size, flags, raw_ptr, ptr))


def iter_reflexive_offs(map_data, refl_offset, struct_size,
                        max_count=0xFFffFFff, tag_magic=None):
    count, start, _ = read_reflexive(map_data, refl_offset, max_count,
                                     struct_size, tag_magic)
    return range(start, start + count*struct_size, struct_size)


def repair_dependency(index_array, map_data, tag_magic, repair, engine, cls,
                      dep_offset, map_magic=None):
    if map_magic is None:
        map_magic = tag_magic

    dep_offset -= tag_magic

    try:
        map_data.seek(dep_offset + 12)
    except ValueError:
        return

    tag_id = int.from_bytes(map_data.read(4), "little")
    tag_index_id = tag_id & 0xFFff

    if tag_index_id not in range(len(index_array)):
        return

    tag_index_ref = index_array[tag_index_id]
    if tag_index_ref.id != tag_id:
        return

    if cls is None:
        map_data.seek(dep_offset)
        cls = map_data.read(4)

    # if the class is obje or shdr, make sure to get the ACTUAL class
    if cls in (b'ejbo', b'meti', b'ived', b'tinu'):
        map_data.seek(tag_index_ref.meta_offset - map_magic)
        object_type = int.from_bytes(map_data.read(2), 'little')
        if object_type not in range(-1, len(object_class_bytes) - 1):
            return

        cls = object_class_bytes[object_type]
    elif cls == b'rdhs':
        map_data.seek(tag_index_ref.meta_offset - map_magic + 36)
        shader_type = int.from_bytes(map_data.read(2), 'little')
        if shader_type not in range(3, len(shader_class_bytes) - 1):
            return

        cls = shader_class_bytes[shader_type]
    elif cls in (b'2dom', b'edom'):
        if "xbox" in engine:
            cls = b'edom'
        else:
            cls = b'2dom'

    map_data.seek(dep_offset)
    map_data.write(cls)

    if tag_index_id not in repair:
        repair[tag_index_id] = cls[::-1].decode('latin1')
        #DEBUG
        # print("        %s %s %s" % (tag_id, cls, dep_offset))


def repair_dependency_array(index_array, map_data, magic, repair, engine,
                            base_class, start, array_size, struct_size=16,
                            map_magic=None):
    '''Macro for deprotecting a contiguous array of dependencies'''
    for offset in range(start, start + array_size*struct_size, struct_size):
        repair_dependency(
            index_array, map_data, magic, repair, engine, base_class, offset,
            map_magic)


class_bytes_by_fcc = {
    "senv": b'vnes' + b'rdhs' + NULL_CLASS,  # 3
    "soso": b'osos' + b'rdhs' + NULL_CLASS,  # 4
    "sotr": b'rtos' + b'rdhs' + NULL_CLASS,  # 5
    "schi": b'ihcs' + b'rdhs' + NULL_CLASS,  # 6
    "scex": b'xecs' + b'rdhs' + NULL_CLASS,  # 7
    "swat": b'taws' + b'rdhs' + NULL_CLASS,  # 8
    "sgla": b'algs' + b'rdhs' + NULL_CLASS,  # 9
    "smet": b'tems' + b'rdhs' + NULL_CLASS,  # 10
    "spla": b'alps' + b'rdhs' + NULL_CLASS,  # 11
    "shdr": b'rdhs' + NULL_CLASS + NULL_CLASS,  # -1

    "bipd": b'dpib' + b'tinu' + b'ejbo',  # 0
    "vehi": b'ihev' + b'tinu' + b'ejbo',  # 1
    "weap": b'paew' + b'meti' + b'ejbo',  # 2
    "eqip": b'piqe' + b'meti' + b'ejbo',  # 3
    "garb": b'brag' + b'meti' + b'ejbo',  # 4
    "proj": b'jorp' + b'ejbo' + NULL_CLASS,  # 5
    "scen": b'necs' + b'ejbo' + NULL_CLASS,  # 6
    "mach": b'hcam' + b'ived' + b'ejbo',  # 7
    "ctrl": b'lrtc' + b'ived' + b'ejbo',  # 8
    "lifi": b'ifil' + b'ived' + b'ejbo',  # 9
    "plac": b'calp' + b'ejbo' + NULL_CLASS,  # 10
    "ssce": b'ecss' + b'ejbo' + NULL_CLASS,  # 11
    "obje": b'ejbo' + NULL_CLASS + NULL_CLASS  # -1
    }

for cls in tag_class_be_int_to_fcc_os.values():
    if cls not in class_bytes_by_fcc:
        cls_1 = bytes(cls[slice(None, None, -1)], "latin1")
        cls_2 = cls_3 = NULL_CLASS
        if cls_1 == b'gpfe':
            cls_2 == b'ppfe'
        elif cls_1 == b'gphs':
            cls_2 == b'pphs'

        class_bytes_by_fcc[cls] = cls_1 + cls_2 + cls_3



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: DarkShallFall
# 	Mapped out and named reflexives and values
# revision: 3		author: Lord Zedd
# 	Updates.
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


coll_material = Struct("material",
    h3_string_id("name"),
    ENDIAN=">", SIZE=4
    )


coll_region_permutation_bsp_bsp_3d_node = Struct("bsp_3d_node",
    UInt8("unknown_0", VISIBLE=False),
    SInt16("second_child"),
    UInt8("unknown_1", VISIBLE=False),
    SInt16("first_child"),
    SInt16("plane"),
    ENDIAN=">", SIZE=8
    )


coll_region_permutation_bsp_plane = Struct("plane",
    QStruct("plane", INCLUDE=ijk_float),
    Float("plane_d"),
    ENDIAN=">", SIZE=16
    )


coll_region_permutation_bsp_leave = Struct("leave",
    SInt16("flags"),
    SInt16("bsp_2d_reference_count"),
    SInt16("unknown", VISIBLE=False),
    SInt16("first_bsp_2d_reference"),
    ENDIAN=">", SIZE=8
    )


coll_region_permutation_bsp_bsp_2d_reference = Struct("bsp_2d_reference",
    SInt16("plane"),
    SInt16("bsp_2d_node"),
    ENDIAN=">", SIZE=4
    )


coll_region_permutation_bsp_bsp_2d_node = Struct("bsp_2d_node",
    QStruct("plane", INCLUDE=ij_float),
    Float("plane_d"),
    SInt16("left_child"),
    SInt16("right_child"),
    ENDIAN=">", SIZE=16
    )


coll_region_permutation_bsp_surface = Struct("surface",
    SInt16("plane"),
    SInt16("first_edge"),
    SInt16("material"),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("breakable_surface"),
    SInt16("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=12
    )


coll_region_permutation_bsp_edge = Struct("edge",
    SInt16("start_vertex"),
    SInt16("end_vertex"),
    SInt16("forward_edge"),
    SInt16("reverse_edge"),
    SInt16("left_surface"),
    SInt16("right_surface"),
    ENDIAN=">", SIZE=12
    )


coll_region_permutation_bsp_vertice = Struct("vertice",
    Pad(12),
    SInt16("first_edge"),
    SInt16("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=16
    )


coll_region_permutation_bsp = Struct("bsp",
    SInt16("node_index"),
    SInt16("unknown", VISIBLE=False),
    h3_reflexive("bsp_3d_nodes", coll_region_permutation_bsp_bsp_3d_node),
    h3_reflexive("planes", coll_region_permutation_bsp_plane),
    h3_reflexive("leaves", coll_region_permutation_bsp_leave),
    h3_reflexive("bsp_2d_references", coll_region_permutation_bsp_bsp_2d_reference),
    h3_reflexive("bsp_2d_nodes", coll_region_permutation_bsp_bsp_2d_node),
    h3_reflexive("surfaces", coll_region_permutation_bsp_surface),
    h3_reflexive("edges", coll_region_permutation_bsp_edge),
    h3_reflexive("vertices", coll_region_permutation_bsp_vertice),
    ENDIAN=">", SIZE=100
    )


coll_region_permutation_bsp_physic = Struct("bsp_physic",
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    SInt16("size_0"),
    SInt16("size_1"),
    SInt32("offset_0"),
    SInt32("unknown_1", VISIBLE=False),
    BytesRaw("unknown_2", SIZE=16, VISIBLE=False),
    Float("unknown_3"),
    Float("unknown_4"),
    Float("unknown_5"),
    BytesRaw("unknown_6", SIZE=4, VISIBLE=False),
    Float("unknown_7"),
    Float("unknown_8"),
    Float("unknown_9"),
    BytesRaw("unknown_10", SIZE=4, VISIBLE=False),
    dependency_uint32("model", VISIBLE=False),
    BytesRaw("unknown_11", SIZE=4, VISIBLE=False),
    SInt16("unknown_12", VISIBLE=False),
    SInt16("unknown_13", VISIBLE=False),
    BytesRaw("unknown_14", SIZE=8, VISIBLE=False),
    SInt16("size_2"),
    SInt16("size_3"),
    SInt32("offset_1"),
    SInt32("unknown_15", VISIBLE=False),
    SInt32("unknown_16", VISIBLE=False),
    BytesRaw("unknown_17", SIZE=12, VISIBLE=False),
    ENDIAN=">", SIZE=112
    )


coll_region_permutation_bsp_mopp_code_data = Struct("data",
    UInt8("data_byte"),
    ENDIAN=">", SIZE=1
    )


coll_region_permutation_bsp_mopp_code = Struct("bsp_mopp_code",
    SInt32("size_0"),
    SInt32("size_1"),
    SInt32("offset_0"),
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    QStruct("offset_1", INCLUDE=xyz_float),
    Float("offset_scale"),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    SInt32("data_size"),
    UInt32("data_capacity"),
    Array("unknown_array", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    h3_reflexive("data", coll_region_permutation_bsp_mopp_code_data),
    BytesRaw("unknown_2", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=64
    )


coll_region_permutation = Struct("permutation",
    h3_string_id("name"),
    h3_reflexive("bsps", coll_region_permutation_bsp),
    h3_reflexive("bsp_physics", coll_region_permutation_bsp_physic),
    h3_reflexive("bsp_mopp_codes", coll_region_permutation_bsp_mopp_code),
    ENDIAN=">", SIZE=40
    )


coll_region = Struct("region",
    h3_string_id("name"),
    h3_reflexive("permutations", coll_region_permutation),
    ENDIAN=">", SIZE=16
    )


coll_pathfinding_sphere = Struct("pathfinding_sphere",
    SInt16("node"),
    Bool16("flags",
        "remains_when_open",
        "vehicle_only",
        "with_sectors",
        ),
    QStruct("center", INCLUDE=xyz_float),
    Float("radius"),
    ENDIAN=">", SIZE=20
    )


coll_node = Struct("node",
    h3_string_id("name"),
    SInt16("unknown"),
    SInt16("parent_node"),
    SInt16("next_sibling_node"),
    SInt16("first_child_node"),
    ENDIAN=">", SIZE=12
    )


coll_body = Struct("tagdata",
    SInt32("collision_model_checksum"),
    BytesRaw("unknown", SIZE=12, VISIBLE=False),
    Bool32("flags",
        "contains_open_edges",
        ),
    h3_reflexive("materials", coll_material),
    h3_reflexive("regions", coll_region),
    h3_reflexive("pathfinding_spheres", coll_pathfinding_sphere),
    h3_reflexive("nodes", coll_node),
    ENDIAN=">", SIZE=68
    )


def get():
    return coll_def

coll_def = TagDef("coll",
    h3_blam_header('coll'),
    coll_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["coll"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os
from collections import OrderedDict
from copy import deepcopy
from datetime import datetime
from xml.etree.ElementTree import ElementTree
from traceback import format_exc
from string import ascii_letters
from supyr_struct.util import str_to_identifier


VALID_MODULE_NAME_CHARS = ascii_letters + '_' + '0123456789'


# Need to fix bug with reflexives sometimes sharing names with earlier fields
#     Occurs in lswd, chgd, and sbsp


engine_specific_field_types = set((
    "string_id", "rawdata_ref", "dependency", "reflexive"
    ))

name_only_field_types = set((
    "string_id", "rawdata_ref", "dependency", "dependency_uint32",
    "float_rad", "yp_float_rad", "ypr_float_rad",
    "float_deg", "yp_float_deg", "ypr_float_deg",
    "color_argb_float", "color_argb_uint32",
    "color_rgb_float",  "color_xrgb_uint32",
    "Float", "from_to_rad", "ascii_str32",
    "SInt8", "SInt16", "SInt32", "UInt8", "UInt16", "UInt32",
    ))

do_not_indent_field_types = set(name_only_field_types)
do_not_indent_field_types.add("reflexive")
do_not_indent_field_types.add("BytesRaw")
do_not_indent_field_types.add("StrLatin1")
do_not_indent_field_types.add("StrUTF16")
do_not_indent_field_types.add("Pad")
do_not_indent_field_types.add("string_id")

string_field_types = set((
    "BytesRaw", "StrLatin1", "StrUTF16",
    ))

enum_field_types = set((
    "SEnum8", "SEnum16", "SEnum32", "UEnum8", "UEnum16", "UEnum32"
    ))

bool_field_types = set((
    "Bool8", "Bool16", "Bool32"
    ))


array_allowed_fields = set((
    "color_argb_float", "color_argb_uint32",
    "color_rgb_float",  "color_xrgb_uint32",
    "Float", "float_rad", "string_id",
    "SInt8", "SInt16", "SInt32",
    "UInt8", "UInt16", "UInt32", "Pad",
    "ascii_str32"
    ))


field_sizes = {
    "reflexive": 12, "rawdata_ref": 20, "dependency": 16,
    "color_argb_float": 16, "color_argb_uint32": 4,
    "color_rgb_float": 12,  "color_xrgb_uint32": 4,
    "Float": 4, "float_rad": 4, "string_id": 4,
    "SInt8":  1, "SInt16":  2, "SInt32":  4,
    "UInt8":  1, "UInt16":  2, "UInt32":  4,
    "SEnum8": 1, "SEnum16": 2, "SEnum32": 4,
    "UEnum8": 1, "UEnum16": 2, "UEnum32": 4,
    "Bool8":  1, "Bool16":  2, "Bool32":  4,
    "ascii_str32": 32, "dependency_uint32": 4,

    # special values
    "bit": 1, "opt": 1, "Pad": 4
    }

type_name_map = {
    "raw": "BytesRaw", "undefined": "Pad", "array": "Array",
    "reflexive": "reflexive", "dataref": "rawdata_ref",
    "tagref": "dependency", "tagref_uint32": "dependency_uint32",
    "float32": "Float", "float": "Float", "degree": "float_rad",
    "int8": "SInt8",  "int16": "SInt16",  "int32": "SInt32",
    "uint8": "UInt8", "uint16": "UInt16", "uint32": "UInt32",
    "enum8": "SEnum8",  "enum16": "SEnum16",  "enum32": "SEnum32",
    "uenum8": "UEnum8", "uenum16": "UEnum16", "uenum32": "UEnum32",
    "bitfield8": "Bool8", "bitfield16": "Bool16", "bitfield32": "Bool32",
    "ascii": "StrLatin1", "utf16": "StrUTF16",
    "stringid": "string_id",
    "colorf": "color_argb_float", "color32": "color_argb_uint32",
    "colorfnoalpha": "color_rgb_float", "color24": "color_xrgb_uint32",

    "shader": "UInt32", "string": "UInt32", "uniclist": "UInt32",

    # special values
    "bit": "bit", "option": "opt", "plugin": "BlockDef"
    }

ignored_names = set((
    "comment", "revision"
    ))

name_fix_replacements = dict(
    index="idx", count="size", insert="ins", parent="ancestor"
    )

prefix_chains = [
    (("min_", "minimum_"), ("max_", "maximum_",)),
    (("eular_angle_x_", "x_", ), ("y_", ), ("z_", )),
    ]

suffix_chains = [
    (("_min", "_minimum"), ("_max", "_maximum",)),
    (("_u", "u"), ("_v", "v"), ("_w", "w")),
    (("_y", "y", "yaw"), ("_p", "p", "pitch"), ("_r", "r", "roll")),
    (("_x", "x"), ("_y", "y"), ("_z", "z"), ("_w", "w")),
    (("_i", "i"), ("_j", "j"), ("_k", "k"), ("_w", "w")),
    (("_a", "_alpha", "alpha"), ("_r", "_red", "red"),
     ("_g", "_green", "green"), ("_b", "_blue", "blue")),
    (("_r", "_red", "red"), ("_g", "_green", "green"),
     ("_b", "_blue", "blue")),
    ]


class StructNode(list):
    typ = ""
    name = "unknown"
    size = 0
    offset = 0
    visible = True
    value = 0

    desc_kw = ()


def fix_name_identifier(name):
    orig_name = name
    name = name.replace("don't", "dont")
    if name[: 1] == "-" and name[1: 2] != "-":
        name = "neg_" + name[1: ]

    if name[: 1] in "0123456789":
        name = "_" + name

    name = str_to_identifier(name)
    return name_fix_replacements.get(name, name)


def replace_struct_node_section(struct_node, start, count, typ, size, name,
                                report_optimize=True, desc_kw=None):
    new_node = StructNode()
    new_node.typ = typ
    new_node.size = size
    new_node.name = fix_name_identifier(name).rstrip("_")
    new_node.desc_kw = desc_kw
    new_node.offset = struct_node[start].offset
    new_node.visible = struct_node[start].visible
    if report_optimize:
        print("    Replacing %s '%s' with '%s'" %
              (count, struct_node[start].typ, typ))
    struct_node[start: start + count] = [new_node]


def optimize_common_structs(struct_node, engine_name, report_optimize):
    if (struct_node.typ in bool_field_types or
        struct_node.typ in enum_field_types):
        return

    start = 0
    while start < len(struct_node):
        # keep checking fields in the node until it's finished

        field_names_str = node_base_name = new_typ = ""
        field_ct = 0
        check_suffix = check_prefix = True
        while start + field_ct < len(struct_node):
            field = struct_node[start + field_ct]
            name = field.name.lower()
            suffix = prefix = true_suffix = true_prefix = ""
            curr_node_base_name = ""

            for prefix_chain in prefix_chains:
                if not check_prefix: break
                curr_prefix_chain_str = curr_prefix = curr_true_prefix = ""
                for i in range(len(prefix_chain)):
                    curr_prefix = prefix_chain[i][0]
                    # detect the prefix
                    curr_true_prefix = ""
                    for s in prefix_chain[i]:
                        if name.startswith(s):
                            curr_true_prefix = s
                            break

                    if curr_true_prefix:
                        # found the prefix this field ends with
                        break
                    else:
                        # didnt find it. add the current prefix to the chain
                        curr_prefix_chain_str += curr_prefix

                if curr_prefix_chain_str == field_names_str:
                    prefix = curr_prefix
                    true_prefix = curr_true_prefix
                    break

            for suffix_chain in suffix_chains:
                if not check_suffix: break
                curr_suffix_chain_str = curr_suffix = curr_true_suffix = ""
                for i in range(len(suffix_chain)):
                    curr_suffix = suffix_chain[i][0]
                    # detect the suffix
                    curr_true_suffix = ""
                    for s in suffix_chain[i]:
                        if name.endswith(s):
                            curr_true_suffix = s
                            break

                    if curr_true_suffix:
                        # found the suffix this field ends with
                        break
                    else:
                        # didnt find it. add the current suffix to the chain
                        curr_suffix_chain_str += curr_suffix

                if curr_suffix_chain_str == field_names_str:
                    suffix = curr_suffix
                    true_suffix = curr_true_suffix
                    break

            if true_suffix and check_suffix:
                field_ct += 1
                curr_node_base_name = name[: -len(true_suffix)]
                check_prefix = False
            elif true_prefix and check_prefix:
                field_ct += 1
                curr_node_base_name = name[len(true_prefix): ]
                check_suffix = False
            else:
                break

            if field_ct == 1:
                node_base_name = curr_node_base_name

            if curr_node_base_name != node_base_name:
                break
            elif check_prefix:
                field_names_str += prefix
            elif check_suffix:
                field_names_str += suffix

        typ_check = "Float"
        optimize = True
        new_typ = "QStruct"
        new_desc_kw = OrderedDict()
        new_size = 0
        field_names_str = field_names_str.strip("_")
        field_typ = struct_node[start].typ
        if field_names_str == "a_r_g_b":
            if not node_base_name: node_base_name = "color"

            if field_typ == "Float":
                new_typ = "color_argb_float"
                new_size = 16
            elif field_typ == "UInt8":
                new_typ = "color_argb_uint32"
                typ_check = "UInt8"
                new_size = 4
            else:
                optimize = False
        elif field_names_str == "r_g_b":
            if not node_base_name: node_base_name = "color"

            if field_typ == "Float":
                new_typ = "color_rgb_float"
                new_size = 12
            elif field_typ == "UInt8":
                new_typ = "color_xrgb_uint32"
                typ_check = "UInt8"
                new_size = 4
            else:
                optimize = False
        elif field_names_str == "eular_angle_x_y_z":
            if not node_base_name: node_base_name = "rotation"

            if field_typ == "float_rad":
                new_typ = "ypr_float_rad"
                typ_check = "float_rad"
            else:
                optimize = False
        elif field_names_str == "x_y_z":
            if not node_base_name: node_base_name = "position"

            if field_typ == "Float":
                new_desc_kw["INCLUDE"] = "xyz_float"
                new_size = 12
            else:
                optimize = False
        elif field_names_str == "x_y":
            if not node_base_name: node_base_name = "position"

            if field_typ == "Float":
                new_desc_kw["INCLUDE"] = "xy_float"
                new_size = 8
            else:
                optimize = False
        elif field_names_str in ("i_j_k_w", "x_y_z_w"):
            if not node_base_name: node_base_name = "rotation"

            if field_typ == "Float":
                new_desc_kw["INCLUDE"] = "ijkw_float"
                new_size = 16
            elif field_typ == "SInt16":
                new_desc_kw["INCLUDE"] = "ijkw_sint16"
                typ_check = "SInt16"
                new_size = 8
            else:
                optimize = False
        elif field_names_str == "i_j_k":
            if not node_base_name: node_base_name = "rotation"

            if field_typ == "Float":
                new_desc_kw["INCLUDE"] = "ijk_float"
                new_size = 12
            elif field_typ == "float_rad":
                new_typ = "ypr_float_rad"
                typ_check = "float_rad"
            elif field_typ == "SInt16":
                new_desc_kw["INCLUDE"] = "ijk_sint16"
                typ_check = "SInt16"
                new_size = 6
            else:
                optimize = False
        elif field_names_str == "i_j":
            if not node_base_name: node_base_name = "angle"

            new_size = 8
            if field_typ == "Float":
                new_desc_kw["INCLUDE"] = "ij_float"
            elif field_typ == "float_rad":
                new_typ = "yp_float_rad"
                typ_check = "float_rad"
            elif field_typ == "SInt16":
                new_desc_kw["INCLUDE"] = "ij_sint16"
                typ_check = "SInt16"
                new_size = 4
            else:
                optimize = False
        elif field_names_str == "u_v":
            if not node_base_name: node_base_name = "tex position"

            if field_typ == "Float":
                new_desc_kw["INCLUDE"] = "uv_float"
                new_size = 8
            else:
                optimize = False
        elif field_names_str == "y_p_r":
            if not node_base_name: node_base_name = "angle"

            new_size = 12
            if field_typ == "Float":
                new_typ = "ypr_float_deg"
            elif field_typ == "float_rad":
                new_typ = "ypr_float_rad"
                typ_check = "float_rad"
            else:
                optimize = False
        elif field_names_str == "y_p":
            if not node_base_name: node_base_name = "angle"

            new_size = 8
            if field_typ == "Float":
                new_typ = "yp_float_deg"
            elif field_typ == "float_rad":
                new_typ = "yp_float_rad"
                typ_check = "float_rad"
            else:
                optimize = False
        elif field_names_str == "min_max":
            if not node_base_name: node_base_name = "bounds"

            new_size = 8
            if field_typ == "Float":
                new_desc_kw["INCLUDE"] = "from_to"
            elif field_typ == "float_rad":
                new_typ = "from_to_rad"
                typ_check = "float_rad"
            elif field_typ[1: ] in ("Int32", "Int16", "Int8"):
                new_desc_kw["INCLUDE"] = "from_to_" + field_typ.lower()
                typ_check = field_typ
                if field_typ in ("UInt16", "SInt16"):
                    new_size = 4
                elif field_typ in ("UInt8", "SInt8"):
                    new_size = 2
            else:
                optimize = False
        else:
            optimize = False

        if optimize:
            for i in range(start, start + field_ct):
                if struct_node[i].typ != typ_check:
                    optimize = False
                    break

        if optimize:
            replace_struct_node_section(struct_node, start, field_ct, new_typ,
                                        new_size, node_base_name, report_optimize,
                                        new_desc_kw)

        start += 1


def optimize_numbered_arrays(struct_node, engine_name, report_optimize):
    if (struct_node.typ in bool_field_types or
        struct_node.typ in enum_field_types):
        return

    start = 0
    while start < len(struct_node):
        base_field_num = field_num = field_num_index = -1
        base_field_name = field_name = first_field_name = ""
        base_field_typ = ""
        field_ct = 0
        while start + field_ct < len(struct_node):
            field = struct_node[start + field_ct]
            if field.typ not in array_allowed_fields:
                break

            name = field.name.lower()
            name_pieces = field.name.split("_")
            if field_ct == 0:
                while field_num_index != len(name_pieces):
                    try:
                        int(name_pieces[field_num_index])
                        break
                    except Exception:
                        field_num_index += 1

            if field_num_index == len(name_pieces):
                # handle cases where the fields are the same name
                if not first_field_name:
                    first_field_name = base_field_name = name
                    base_field_typ = field.typ

                if first_field_name == name and base_field_typ == field.typ:
                    field_ct += 1
                else:
                    break
                continue

            try:
                field_num = int(name_pieces.pop(field_num_index))
                field_name = "_".join(s for s in name_pieces)
                if not field_name:
                    pass
                elif field_num_index == 0:
                    field_name = "_" + field_name
                elif field_num_index == -1:
                    field_name = field_name + "_"

                if field_ct == 0:
                    base_field_num = field_num
                    base_field_name = field_name
                    base_field_typ = field.typ

                if (field_num  != base_field_num + field_ct or
                    field.typ  != base_field_typ or
                    field_name != base_field_name):
                    break
            except Exception:
                break

            field_ct += 1

        base_field_name = base_field_name.strip("_")
        if field_ct > 3 and base_field_typ in array_allowed_fields:
            new_desc_kw = OrderedDict()
            sub_struct_string = '%s("%s")' % (base_field_typ, base_field_name)
            if base_field_typ in engine_specific_field_types:
                sub_struct_string = "%s_%s" % (engine_name, sub_struct_string)

            new_desc_kw.update(SIZE=field_ct, SUB_STRUCT=sub_struct_string)
            replace_struct_node_section(struct_node, start, field_ct, "Array",
                                        field_ct, base_field_name + "_array",
                                        report_optimize, new_desc_kw)

        start += 1


def optimize_struct_node(struct_node, engine_name, report_optimize):
    optimize_common_structs(struct_node, engine_name, report_optimize)
    optimize_numbered_arrays(struct_node, engine_name, report_optimize)


def struct_node_to_supyr_desc(struct_node, descs_by_name,
                              enum_bool_names_by_desc,
                              shared_enum_bool_names_by_desc,
                              engine_name, parent_name="", indent=0,
                              report_optimize=True):
    optimize_struct_node(struct_node, engine_name, report_optimize)

    added_names = dict()
    add_num_to = set()
    for sub_node in struct_node:
        # take care of fields with the same name
        added_names[sub_node.name] = added_names.get(
            sub_node.name, 0) + 1
        if added_names[sub_node.name] > 1:
            add_num_to.add(sub_node.name)
            sub_node.name += "_%s" % (added_names[sub_node.name] - 1)

    if add_num_to:
        for sub_node in struct_node:
            if sub_node.name not in add_num_to:
                continue
            added = 0
            while "%s_%s" % (sub_node.name, added) in add_num_to:
                added += 1
            sub_node.name = "%s_%s" % (sub_node.name, added)


    desc_name = struct_node_name = struct_node.name
    if struct_node.typ == "reflexive":
        if desc_name.endswith("s"):
            desc_name = struct_node_name = desc_name[: -1]
        if parent_name:
            desc_name = "%s_%s" % (parent_name, desc_name)

    if struct_node.typ == "BlockDef":
        indent = 1

    indent_str = ' ' * (4 * indent)
    if struct_node.typ in ("reflexive", "BlockDef"):
        desc_str = 'Struct('
    else:
        if struct_node.typ in engine_specific_field_types:
            desc_str = "%s%s_%s(" % (indent_str, engine_name, struct_node.typ)
        else:
            desc_str = '%s%s(' % (indent_str, struct_node.typ)

    if struct_node.typ == "Pad":
        desc_str += str(struct_node.size)
        indent_str = ""
    else:
        if struct_node.typ == "BlockDef":
            desc_str += '"tagdata", '
        else:
            desc_str += '"%s", ' % struct_node_name

        if struct_node.typ in name_only_field_types:
            indent_str = ""
        else:
            if struct_node.typ in string_field_types:
                desc_str += "SIZE=%s, " % struct_node.size
                indent_str = ""

            if struct_node.typ in enum_field_types:
                i = 0
                last_val = -1
                enum_str = ""
                for enum in struct_node:
                    if enum.value == i or (enum.value - last_val == 1):
                        enum_str += '%s"%s",\n' % (
                            indent_str, enum.name)
                    else:
                        enum_str += '%s("%s", %s),\n' % (
                            indent_str, enum.name, enum.value)
                    i += 1
                    last_val = enum.value

                enum_str = "(\n%s%s)" % (enum_str, indent_str)
                enum_name = "%s_%s" % (parent_name, desc_name)
                if enum_str in enum_bool_names_by_desc:
                    enum_name = enum_bool_names_by_desc[enum_str]
                elif enum_str in shared_enum_bool_names_by_desc:
                    enum_name = shared_enum_bool_names_by_desc[enum_str]
                else:
                    enum_bool_names_by_desc[enum_str] = enum_name

                desc_str += "*%s, " % enum_name
                indent_str = ""

            elif struct_node.typ in bool_field_types:
                has_incremental_flags = True
                base_flag_num = flag_num = -1
                base_flag_name = flag_name = ""
                for bit in struct_node:
                    name_pieces = bit.name.split("_")
                    try:
                        flag_num = int(name_pieces[-1])
                        flag_name = "_".join(s for s in name_pieces[: -1])
                        if flag_name:
                            flag_name += "_"

                        if bit.offset == 0:
                            base_flag_num = flag_num
                            base_flag_name = flag_name

                        if (flag_num != base_flag_num + bit.offset or
                            flag_name != base_flag_name):
                            has_incremental_flags = False
                            break
                    except Exception:
                        has_incremental_flags = False
                        break

                if has_incremental_flags:
                    bits = len(struct_node)
                    if base_flag_name == "":
                        base_name = struct_node.name
                        if base_name.endswith("s"):
                            base_name = base_name[: -1]

                        bools_str = 'tuple("%s_%%s" %% i for i in range(%s))' % (
                            base_name, bits)
                        bools_name = "%s_bits" % desc_name
                        if (bools_str not in enum_bool_names_by_desc and
                            bools_str not in shared_enum_bool_names_by_desc):
                            enum_bool_names_by_desc[bools_str] = bools_name
                    elif base_flag_name == "bit_":
                        bools_str = 'tuple("bit_%%s" %% i for i in range(%s))' % bits
                        bools_name = "unknown_flags_%s" % bits
                        if (bools_str not in enum_bool_names_by_desc and
                            bools_str not in shared_enum_bool_names_by_desc):
                            enum_bool_names_by_desc[bools_str] = bools_name
                    elif base_flag_num == 0:
                        bools_name = '("%s%%s" %% i for i in range(%s))' % (
                            base_flag_name, bits)
                    else:
                        bools_name = '("%s%%s" %% i for i in range(%s, %s))' % (
                            base_flag_name, base_flag_num, base_flag_num + bits)

                    desc_str += "*%s, " % bools_name
                    indent_str = ""
                else:
                    desc_str += '\n'
                    i = 0
                    j = last_val = -1
                    for bit in struct_node:
                        j += 1
                        name_parts = bit.name.split("_")
                        if len(name_parts) == 2 and name_parts[0] == "bit":
                            try:
                                int(name_parts[1])
                                continue
                            except Exception:
                                pass

                        if bit.value == i or (bit.value - last_val == 1):
                            desc_str += '%s%s"%s",\n' % (
                                indent_str, indent_str, bit.name)
                        else:
                            desc_str += '%s%s("%s", 1 << %s),\n' % (
                                indent_str, indent_str,
                                bit.name, bit.value)
                        i += 1
                        last_val = bit.value
                    indent_str = ' ' * (4 * (indent + 1))


            elif struct_node.typ in ("BlockDef", "reflexive",
                                     "Struct", "QStruct", "Array"):

                if len(struct_node):
                    desc_str += '\n'
                elif struct_node.desc_kw:
                    for k, v in struct_node.desc_kw.items():
                        desc_str += "%s=%s, " % (k, v)
                    indent_str = ""

                for field in struct_node:
                    subdesc_str = struct_node_to_supyr_desc(
                        field, descs_by_name, enum_bool_names_by_desc,
                        shared_enum_bool_names_by_desc, engine_name,
                        desc_name, 1, report_optimize)

                    if field.typ == "reflexive":
                        desc_str += '    %s_reflexive("%s", %s)' % (
                            engine_name, field.name, subdesc_str)
                    else:
                        desc_str += subdesc_str

                    desc_str += ",\n"

        if not struct_node.visible:
            add_newline = desc_str.endswith("\n")
            desc_str += indent_str + "VISIBLE=False,"
            if add_newline:
                desc_str += "\n"


    if struct_node.typ in ("reflexive", "BlockDef"):
        desc_str += indent_str
        if struct_node.typ == "BlockDef":
            desc_name += "_meta_def"

        desc_str += 'ENDIAN=">", SIZE=%s\n' % struct_node.size

    desc_str = desc_str.rstrip(", ") + indent_str + ')'

    if struct_node.typ in ("reflexive", "BlockDef"):
        descs_by_name[desc_name] = desc_str
        return desc_name

    return desc_str


def parse_xml_node(xml_node, version_infos=None):
    xml_tag = xml_node.tag.lower()
    if version_infos is None:
        version_infos = []

    if xml_tag == "revisions":
        for xml_subnode in xml_node:
            xml_attribs = {key.lower(): xml_subnode.attrib[key]
                           for key in xml_subnode.attrib}
            author = xml_attribs.get('author', "unspecified")
            version = xml_attribs.get('version', "1")
            comment = ""
            for comment in xml_subnode.itertext(): break
            version_infos.append("revision: %s\t\tauthor: %s" % (version, author))
            version_infos.append("\t%s" % comment)

        version_infos.append("revision: %s\t\tauthor: Moses_of_Egypt" % (int(version) + 1, ))
        version_infos.append("\tCleaned up and converted to SuPyr definition")
        return None
    elif xml_tag in ignored_names:
        return None

    new_node = StructNode()
    xml_attribs = {key.lower(): xml_node.attrib[key]
                   for key in xml_node.attrib}
    xml_tag = xml_tag.replace("colour", "color")

    new_node.name = fix_name_identifier(
        xml_attribs.get('name', "unnamed").lower()).rstrip("_")
    new_node.offset = eval(xml_attribs.get('offset', "None"))
    new_node.visible = eval(xml_attribs.get('visible', "true").capitalize())

    if xml_tag == "tagref" and not eval(xml_attribs.get(
            'withclass', "true").capitalize()):
        xml_tag = "tagref_uint32"
        new_node.visible = False
    elif xml_tag.startswith("colorf") and xml_attribs.get(
            'format', "rgb") == "rgb":
        xml_tag = "colorfnoalpha"


    if xml_tag in type_name_map:
        new_node.typ = type_name_map[xml_tag]
    else:
        raise TypeError("Unknown field type '%s'" % xml_node.tag)

    if "basesize" in xml_attribs:
        new_node.size = eval(xml_attribs['basesize'])
    elif "size" in xml_attribs:
        new_node.size = eval(xml_attribs['size'])
    elif "entrysize" in xml_attribs:
        new_node.size = eval(xml_attribs['entrysize'])
    elif "length" in xml_attribs:
        new_node.size = eval(xml_attribs['length'])
    else:
        new_node.size = field_sizes.get(new_node.typ, 0)

    if new_node.size == 32 and xml_tag == "ascii":
        new_node.typ = "ascii_str32"

    if "index" in xml_attribs:
        new_node.value = eval(xml_attribs['index'])
    else:
        new_node.value = eval(xml_attribs.get('value', "0"))

    if new_node.offset is None:
        new_node.offset = new_node.value

    sub_nodes_by_offset = {}
    for xml_subnode in xml_node:
        new_sub_node = parse_xml_node(xml_subnode, version_infos)
        if new_sub_node is not None:
            sub_nodes_by_offset[new_sub_node.offset] = new_sub_node

    for off in sorted(sub_nodes_by_offset):
        sub_node = sub_nodes_by_offset[off]
        if sub_node.typ == "color_xrgb_uint32":
            # adjust for 4 byte based colors
            sub_node.offset -= 1
            sub_nodes_by_offset.pop(off)
            if sub_node.offset in sub_nodes_by_offset:
                # change it to an alpha based color
                sub_node.typ = "color_argb_uint32"
            sub_nodes_by_offset[sub_node.offset] = sub_node

    last_off = 0
    last_node = None
    for off in sorted(sub_nodes_by_offset):
        sub_node = sub_nodes_by_offset[off]
        sub_node_size = sub_node.size
        if sub_node.typ == "reflexive":
            sub_node_size = 12

        if "Enum" in new_node.typ or "Bool" in new_node.typ:
            pass
        elif last_node is not None and (sub_node.typ == "Pad" and
                                        last_node.typ == "Pad"):
            # consolidate consecutive padding
            last_node.size += sub_node_size
            last_off = off + sub_node_size
            continue
        elif off - last_off:
            # add missing padding
            pad_node = StructNode()
            pad_node.visible = True
            pad_node.typ = "Pad"
            pad_node.name = "unknown"
            pad_node.size = off - last_off
            pad_node.offset = pad_node.value = last_off
            new_node.append(pad_node)

            if pad_node.size < 0:
                raise ValueError(
                    ("Negative padding size in '%s' at field "
                     "'%s' of type '%s' at offset '%s'") %
                    (new_node.name, sub_node.name,
                     sub_node.typ, sub_node.offset))

        new_node.append(sub_node)
        last_off = off + sub_node_size
        last_node = sub_node

    for sub_node in new_node:
        if sub_node.typ == "Pad" and not sub_node.visible:
            sub_node.typ = "BytesRaw"

    return new_node


def parse_xml(xml_path, version_infos):
    try:
        xml_root = ElementTree().parse(xml_path)
        tag_id = os.path.splitext(os.path.basename(xml_path))[0].strip(".")
        nodes = parse_xml_node(xml_root, version_infos)
        nodes.name = fix_name_identifier(tag_id + ("_" * (4 - len(tag_id))))
    except Exception:
        print(format_exc())
        nodes = None

    return nodes


timestamp = datetime.now().strftime("%Y/%m/%d  %H:%M")
for _, dirs, __ in os.walk("xml/"):
    tag_def_dirs = dict.fromkeys(dirs)
    for tag_def_dir in tag_def_dirs:
        for _, __, files in os.walk(os.path.join("xml", tag_def_dir)):
            tag_def_dirs[tag_def_dir] = tuple(f[: -4] for f in files if
                                              f.lower().endswith(".xml"))
    break


for tag_def_dir in sorted(tag_def_dirs):

    # TODO: Make this more versatile
    tag_engine_cls_name = tag_def_dir.upper() + "Tag"

    defs_dir = os.path.join(tag_def_dir, "defs")
    os.makedirs(defs_dir, exist_ok=True)

    common_descs_filepath = os.path.join(tag_def_dir, "common_descs.py")
    enum_descs_filepath = os.path.join(tag_def_dir, "enums.py")
    if not os.path.exists(common_descs_filepath):
        with open(common_descs_filepath, "w+") as pyf:
            pyf.write('''from reclaimer.common_descs import *\n''')

    # make the all import in the __init__
    with open(os.path.join(defs_dir, "__init__.py"), "w+") as pyf:
        pyf.write('__all__ = (')
        i = 0
        for fname in sorted(tag_def_dirs[tag_def_dir]):
            if i % 8 == 0:
                pyf.write('\n    ')
            module_name = "".join(c if c in VALID_MODULE_NAME_CHARS
                                  else "_" for c in fname)
            module_name += "_" * ((4 - (len(module_name) % 4)) % 4)
            pyf.write('"%s", ' % module_name)
            i += 1
        pyf.write('\n    )\n')


    all_enum_bool_names_by_desc = {}
    shared_enum_bool_names_by_desc = {}
    all_version_infos = {}
    all_tag_struct_nodes = {}

    # convert all xml's into struct node trees
    print("Parsing XML's...")
    for fname in tag_def_dirs[tag_def_dir]:
        xml_path = os.path.join("xml", tag_def_dir, fname + ".xml")
        version_infos = []
        tag_struct_nodes = parse_xml(xml_path, version_infos)
        if tag_struct_nodes is None:
            print('Could not parse: %s' % xml_path)
        else:
            all_version_infos[fname] = version_infos
            all_tag_struct_nodes[fname] = tag_struct_nodes


    print("Collecting shared enums...")
    # for the first pass, we collect all shared enums
    for fname in sorted(all_tag_struct_nodes):
        descs_by_name = OrderedDict()
        enum_bool_names_by_desc = {}
        struct_node_to_supyr_desc(deepcopy(all_tag_struct_nodes[fname]),
                                  descs_by_name, enum_bool_names_by_desc,
                                  shared_enum_bool_names_by_desc,
                                  tag_def_dir, report_optimize=False)
        for desc_str, desc_name in enum_bool_names_by_desc.items():
            if desc_str in all_enum_bool_names_by_desc:
                shared_enum_bool_names_by_desc[desc_str] = desc_name

            all_enum_bool_names_by_desc[desc_str] = desc_name


    # write out the shared enums
    with open(enum_descs_filepath, "w+") as pyf:
        pyf.write('''from reclaimer.enums import *\n''')
        shared_enum_bool_descs_by_name = {
            shared_enum_bool_names_by_desc[desc]: desc for
            desc in shared_enum_bool_names_by_desc}
        for desc_name in sorted(shared_enum_bool_descs_by_name):
            desc_str = shared_enum_bool_descs_by_name[desc_name]
            pyf.write("\n%s = %s\n" % (desc_name, desc_str.strip(",")))


    print("Writing definitions...")
    # on the second pass, we actually write the definitions
    for tag_cls in sorted(all_tag_struct_nodes):
        print(tag_cls)
        version_infos = all_version_infos[tag_cls]
        tag_struct_nodes = all_tag_struct_nodes[tag_cls]

        descs_by_name = OrderedDict()
        enum_bool_names_by_desc = {}
        struct_node_to_supyr_desc(tag_struct_nodes, descs_by_name,
                                  enum_bool_names_by_desc,
                                  shared_enum_bool_names_by_desc,
                                  tag_def_dir, report_optimize=True)

        module_name = "".join(c if c in VALID_MODULE_NAME_CHARS
                              else "_" for c in tag_cls)
        module_name += "_" * ((4 - (len(module_name) % 4)) % 4)
        with open(os.path.join(defs_dir, module_name) + ".py", "w+") as pyf:
            pyf.write("############# Credits and version info #############\n"
                      "# Definition generated from Assembly XML tag def\n"
                      "#\t Date generated: %s\n#\n" % timestamp)

            for version_info_str in version_infos:
                pyf.write("# %s\n" % version_info_str)

            pyf.write("#\n"
                      "####################################################\n")

            pyf.write('''
from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef''')
            enum_bool_descs_by_name = {enum_bool_names_by_desc[desc]: desc for
                                       desc in enum_bool_names_by_desc}

            for desc_name in sorted(enum_bool_descs_by_name):
                pyf.write("\n\n")
                desc_str = enum_bool_descs_by_name[desc_name]
                pyf.write("%s = %s" % (desc_name, desc_str.strip(",")))

            for desc_name in descs_by_name:
                pyf.write("\n\n\n")
                desc_str = descs_by_name[desc_name]
                if desc_name == "%s_meta_def" % module_name:
                    desc_name = "%s_body" % module_name
                pyf.write("%s = %s" % (desc_name, desc_str.strip(",")))


            pyf.write('\n\n\ndef get():\n    return %s_def' % module_name)

            if "_" in tag_cls:
                print("    MUST SET THIS TAG CLASS STRING MANUALLY")
                pyf.write("\n\n#  REMINDER: Set this tag class string manually.")

            pyf.write('''\n
%s_def = TagDef("%s",
    %s_blam_header('%s'),
    %s_body,

    ext=".%%s" %% %s_tag_class_fcc_to_ext["%s"], endian=">", tag_cls=%s
    )''' % (module_name, tag_cls, tag_def_dir, tag_cls, module_name,
            tag_def_dir, tag_cls, tag_engine_cls_name))

print("Finished")



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Cleanup and porting
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


lens_reflection = Struct("reflection",
    Bool16("flags",
        "align_rotation_with_screen_center",
        "radius_not_scaled_by_distance",
        "radius_scaled_by_occlusion_factor",
        "occluded_by_solid_objects",
        "ignore_light_color",
        "not_affected_by_inner_occlusion",
        ),
    SInt16("bitmap_index"),
    Float("position_along_flare_axis"),
    Float("rotation_offset"),
    QStruct("radius", INCLUDE=from_to),
    QStruct("brightness", INCLUDE=from_to),
    Float("tint_modulation_factor"),
    color_rgb_float("tint_color"),
    Float("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=48
    )


lens_brightnes = Struct("brightnes",
    h3_rawdata_ref("function"),
    ENDIAN=">", SIZE=20
    )


lens_color = Struct("color",
    h3_rawdata_ref("function"),
    ENDIAN=">", SIZE=20
    )


lens_unknown_5 = Struct("unknown_5",
    BytesRaw("unknown", SIZE=16, VISIBLE=False),
    h3_rawdata_ref("function"),
    ENDIAN=">", SIZE=36
    )


lens_rotation = Struct("rotation",
    BytesRaw("unknown", SIZE=16, VISIBLE=False),
    h3_rawdata_ref("function"),
    ENDIAN=">", SIZE=36
    )


lens_body = Struct("tagdata",
    float_rad("falloff_angle"),
    float_rad("cutoff_angle"),
    Float("occlusion_radius"),
    SInt32("unknown_0", VISIBLE=False),
    BytesRaw("unknown_1", SIZE=8, VISIBLE=False),
    Float("near_fade_distance"),
    Float("far_fade_distance"),
    h3_dependency("bitmap"),
    Array("unknown_array", SUB_STRUCT=SInt16("unknown"), SIZE=4, VISIBLE=False),
    float_rad("rotation_function_scale"),
    SInt16("unknown_2", VISIBLE=False),
    SInt16("unknown_3", VISIBLE=False),
    h3_reflexive("reflections", lens_reflection),
    BytesRaw("unknown_4", SIZE=4, VISIBLE=False),
    h3_reflexive("brightness", lens_brightnes),
    h3_reflexive("color", lens_color),
    h3_reflexive("unknown_5", lens_unknown_5),
    h3_reflexive("rotation", lens_rotation),
    BytesRaw("unknown_6", SIZE=24, VISIBLE=False),
    ENDIAN=">", SIZE=152
    )


def get():
    return lens_def

lens_def = TagDef("lens",
    h3_blam_header('lens'),
    lens_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["lens"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.common_descs import *
from reclaimer.meta.objs.halo1_rsrc_map import Halo1RsrcMapTag
from supyr_struct.defs.tag_def import TagDef


def get():
    return halo1_rsrc_map_def


def tag_path_pointer(parent=None, new_value=None, **kwargs):
    if parent is None:
        raise KeyError()
    t_head = parent.parent
    if new_value is None:
        return t_head.parent.parent.tag_paths_pointer + t_head.path_offset
    t_head.path_offset = new_value - t_head.parent.parent.tag_paths_pointer


rsrc_tag = Container("tag",
    BytesRaw("data", SIZE="..size", POINTER="..offset"),
    CStrTagRef("path", POINTER=tag_path_pointer, MAX=768, WIDGET=EntryFrame),
    )

tag_header = Struct("tag header",
    LUInt32("path offset"),
    LUInt32("size"),
    LUInt32("offset"),
    STEPTREE=rsrc_tag
    )

halo1_rsrc_map_def = TagDef("halo1_rsrc_map",
    LUEnum32("resource type",
        'NONE',
        'bitmaps',
        'sounds',
        'strings'
        ),
    LPointer32("tag paths pointer"),
    LPointer32("tag headers pointer"),
    LUInt32("tag count"),
    Array("tags",
        SIZE='.tag_count', SUB_STRUCT=tag_header,
        POINTER='.tag_headers_pointer',
        ),
    endian="<", ext=".map", tag_cls=Halo1RsrcMapTag
    )

lite_rsrc_tag = Container("tag",
    Void("data"),
    CStrLatin1("path", POINTER=tag_path_pointer),
    )
lite_tag_header = Struct("tag header",
    INCLUDE=tag_header, STEPTREE=lite_rsrc_tag
    )
lite_halo1_rsrc_map_desc = dict(halo1_rsrc_map_def.descriptor)
lite_halo1_rsrc_map_desc[4] = Array("tags",
    SIZE='.tag_count', SUB_STRUCT=lite_tag_header,
    POINTER='.tag_headers_pointer',
    )
lite_halo1_rsrc_map_def = TagDef("lite_halo1_rsrc_map",
    descriptor=lite_halo1_rsrc_map_desc,
    endian="<", ext=".map", tag_cls=Halo1RsrcMapTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

def get(): return dobc_def

detail_object_type = Struct("detail_object_type",
    ascii_str32("name"),
    SInt8("sequence_index", SIDETIP="[0,15]"),
    Bool8("scale_flags",
        ("interpolate_color_in_hsv", 4),
        ("more_colors", 8),
        ),
    #UInt8("unknown0", VISIBLE=False),
    Pad(1),
    UInt8("sequence_sprite_count", VISIBLE=False),
    float_zero_to_one("color_override_factor"),
    Pad(8),
    float_wu("near_fade_distance"),
    float_wu("far_fade_distance"),
    Float("size", SIDETIP="world units/pixel"),
    Pad(4),
    QStruct("minimum_color", INCLUDE=rgb_float),
    QStruct("maximum_color", INCLUDE=rgb_float),
    #QStruct("ambient_color", INCLUDE=argb_byte),
    UInt32("ambient_color", INCLUDE=argb_uint32),
    SIZE=96
    )

dobc_body = Struct("tagdata",
    SEnum16("anchor",
        "screen-facing",
        "viewer-facing",
        ),
    Pad(2),
    Float("global_z_offset",
        SIDETIP="applied to all these detail object so they dont float"),
    Pad(44),
    dependency("sprite_plate", "bitm"),
    reflexive("detail_object_types", detail_object_type, 16,
        DYN_NAME_PATH='.name'),
    SIZE=128,
    )

dobc_def = TagDef("dobc",
    blam_header('dobc'),
    dobc_body,

    ext=".detail_object_collection", endian=">", tag_cls=HekTag
    )



try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension

import reclaimer

long_desc = ""
try:
    long_desc = open("README.MD").read()
except Exception:
    print("Couldn't read readme.")


setup(
    name='reclaimer',
    description='A libray of SupyrStruct structures and objects for '
                'games built with the Blam engine',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    version='%s.%s.%s' % reclaimer.__version__,
    url=reclaimer.__website__,
    project_urls={
        #"Documentation": <Need a string entry here>,
        "Source": reclaimer.__website__,
        "Funding": "https://liberapay.com/MEK/",
    },
    author=reclaimer.__author__,
    author_email='MosesBobadilla@gmail.com',
    license='GPLv3',
    packages=[
        'reclaimer',
        'reclaimer.animation',
        'reclaimer.bitmaps',
        'reclaimer.h2',
        'reclaimer.h2.defs',
        'reclaimer.h2.defs.objs',
        'reclaimer.h3',
        'reclaimer.h3.defs',
        'reclaimer.h3.defs.objs',
        'reclaimer.halo_script',
        'reclaimer.hek',
        'reclaimer.hek.defs',
        'reclaimer.hek.defs.objs',
        'reclaimer.meta',
        'reclaimer.meta.gen3_resources',
        'reclaimer.meta.objs',
        'reclaimer.meta.wrappers',
        'reclaimer.meta.wrappers.ext',
        'reclaimer.model',
        'reclaimer.model.jms',
        'reclaimer.misc',
        'reclaimer.misc.defs',
        'reclaimer.misc.defs.objs',
        'reclaimer.os_hek',
        'reclaimer.os_hek.defs',
        'reclaimer.os_hek.defs.objs',
        'reclaimer.os_v3_hek',
        'reclaimer.os_v3_hek.defs',
        'reclaimer.os_v4_hek',
        'reclaimer.os_v4_hek.defs',
        'reclaimer.physics',
        'reclaimer.sounds',
        'reclaimer.sounds.ext',
        'reclaimer.shadowrun_prototype',
        'reclaimer.shadowrun_prototype.defs',
        'reclaimer.strings',
        'reclaimer.stubbs',
        'reclaimer.stubbs.defs',
        'reclaimer.stubbs.defs.objs',
        'reclaimer.util',
        ],
    ext_modules = [
        Extension("reclaimer.sounds.ext.adpcm_ext",
            sources=["reclaimer/sounds/src/adpcm_ext.c",
                     "reclaimer/sounds/src/adpcm-xq/adpcm-lib.c",]),
        Extension("reclaimer.meta.wrappers.ext.byteswapping_ext",
            sources=["reclaimer/meta/wrappers/src/byteswapping_ext.c"]),
        ],
    package_data={
        'reclaimer': [
            "src/*", "meta/wrappers/src/*",
            "sounds/src/*", "sounds/src/adpcm-xq/*",
            '*.[Tt][Xx][Tt]', '*.MD', '*.h',
            '**/p8_palette_halo',   '**/p8_palette_halo_diff_map',
            '**/p8_palette_stubbs', '**/p8_palette_stubbs_diff_map',
            ]
        },
    platforms=["POSIX", "Windows"],
    keywords=["reclaimer", "halo"],
    install_requires=['supyr_struct>=1.5.0', 'binilla', 'arbytmap'],
    requires=['supyr_struct', 'binilla', 'arbytmap'],
    provides=['reclaimer'],
    python_requires=">=3.5",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: C",
        ],
    zip_safe=False,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.hek.defs.objs.tag import HekTag

class Snd_Tag(HekTag):

    def calc_internal_data(self):
        HekTag.calc_internal_data(self)

        for pitch_range in self.data.tagdata.pitch_ranges.STEPTREE:
            pitch_range.playback_rate = 1
            if pitch_range.natural_pitch:
                pitch_range.playback_rate = 1 / pitch_range.natural_pitch



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.unit import *

# replace the grenade types enumerator with an open sauce one
unit_attrs = dict(unit_attrs)
unit_attrs[49] = SEnum16('grenade_type', *grenade_types_os)

unit_body = Struct('tagdata', unit_attrs)

def get():
    return unit_def

unit_def = TagDef("unit",
    blam_header('unit', 2),
    unit_body,

    ext=".unit", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.actr import ActrTag
from supyr_struct.defs.tag_def import TagDef

danger_triggers = (
    "never",
    "visible",
    "shooting",
    "shooting_near_us",
    "damaging_us",
    "unused1",
    "unused2",
    "unused3",
    "unused4",
    "unused5",
    )

actr_body = Struct("tagdata",
    Bool32('flags',
        "can_see_in_darkness",
        "sneak_uncovering_target",
        "sneak_uncovering_pursuit_location",
        "unused1",
        "shoot_at_targets_last_location",
        "try_to_stay_still_when_crouched",
        "crouch_when_not_in_combat",
        "crouch_when_guarding",
        "unused2",
        "must_crouch_to_shoot",
        "panic_when_surprised",
        "always_charge_at_enemies",
        "gets_in_vehicles_with_player",
        "starts_firing_before_aligned",
        "standing_must_move_forward",
        "crouching_must_move_forward",
        "defensive_crouch_while_charging",
        "use_stalking_behavior",
        "stalking_freeze_if_exposed",
        "always_berserk_in_attacking_mode",
        "berserking_uses_panicked_movement",
        "flying",
        "panicked_by_unopposable_enemy",
        "crouch_when_hiding_from_unopposable",
        "always_charge_in_attacking_mode",
        "dive_off_ledges",
        "swarm",
        "suicidal_melee_attack",
        "cannot_move_while_crouching",
        "fixed_crouch_facing",
        "crouch_when_in_line_of_fire",
        "avoid_friends_line_of_fire"
        ),
    Bool32('more_flags',
        "avoid_all_enemy_attack_vectors",
        "must_stand_to_fire",
        "must_stop_to_fire",
        "disallow_vehicle_combat",
        "pathfinding_ignores_danger",
        "panic_in_groups",
        "no_corpse_shooting"
        ),

    Pad(12),
    SEnum16("type", *actor_types),

    Pad(2),
    Struct("perception",
        float_wu("max_vision_distance"),  # world units
        float_rad("central_vision_angle"),  # radians
        float_rad("max_vision_angle"),  # radians

        Pad(4),
        float_rad("peripheral_vision_angle"),  # radians
        float_wu("peripheral_distance"),  # world units

        Pad(4),
        QStruct("standing_gun_offset", INCLUDE=ijk_float),
        QStruct("crouching_gun_offset", INCLUDE=ijk_float),
        float_wu("hearing_distance"),  # world units
        float_zero_to_one("notice_projectile_chance"),
        float_zero_to_one("notice_vehicle_chance"),

        Pad(8),
        float_sec("combat_perception_time",
                  UNIT_SCALE=sec_unit_scale),  # seconds
        float_sec("guard_perception_time",
                  UNIT_SCALE=sec_unit_scale),  # seconds
        float_sec("non_combat_perception_time",
                  UNIT_SCALE=sec_unit_scale),  # seconds

        float_sec("inv_combat_perception_time",
                  UNIT_SCALE=sec_unit_scale, VISIBLE=False),
        float_sec("inv_guard_perception_time",
                  UNIT_SCALE=sec_unit_scale, VISIBLE=False),
        float_sec("inv_non_combat_perception_time",
                  UNIT_SCALE=sec_unit_scale, VISIBLE=False),
        ),

    Pad(8),
    Struct("movement",
        float_zero_to_one("dive_into_cover_chance"),
        float_zero_to_one("emerge_from_cover_chance"),
        float_zero_to_one("dive_from_grenade_cover_chance"),
        float_wu("pathfinding_radius"),  # world units
        float_zero_to_one("glass_ignorance_chance"),
        float_wu("stationary_movement_dist"),  # world units
        float_wu("free_flying_sidestep"),  # world units
        float_rad("begin_moving_angle"),  # radians
        ),

    Pad(4),
    Struct("looking",
        yp_float_rad("maximum_aiming_deviation"),  # radians
        yp_float_rad("maximum_looking_deviation"),  # radians
        float_rad("noncombat_look_delta_l"),  # radians
        float_rad("noncombat_look_delta_r"),  # radians
        float_rad("combat_look_delta_l"),  # radians
        float_rad("combat_look_delta_r"),  # radians
        from_to_rad("idle_aiming_range"),  # radians
        from_to_rad("idle_looking_range"),  # radians
        QStruct("event_look_time_modifier", INCLUDE=from_to),
        from_to_sec("noncombat_idle_facing"),  # seconds
        from_to_sec("noncombat_idle_aiming"),  # seconds
        from_to_sec("noncombat_idle_looking"),  # seconds
        from_to_sec("guard_idle_facing"),  # seconds
        from_to_sec("guard_idle_aiming"),  # seconds
        from_to_sec("guard_idle_looking"),  # seconds
        from_to_sec("combat_idle_facing"),  # seconds
        from_to_sec("combat_idle_aiming"),  # seconds
        from_to_sec("combat_idle_looking"),  # seconds
        Pad(8),
        from_to_neg_one_to_one("cosine_maximum_aiming_deviation", VISIBLE=False),
        from_to_neg_one_to_one("cosine_maximum_looking_deviation", VISIBLE=False),

        dependency("DO_NOT_USE_1", "weap"),

        Pad(268),
        dependency("DO_NOT_USE_2", "proj")
        ),

    Struct("unopposable",
        SEnum16("unreachable_danger_trigger", *danger_triggers),
        SEnum16("vehicle_danger_trigger", *danger_triggers),
        SEnum16("player_danger_trigger", *danger_triggers),

        Pad(2),
        from_to_sec("danger_trigger_time"),  # seconds
        SInt16("friends_killed_trigger"),
        SInt16("friends_retreating_trigger"),

        Pad(12),
        from_to_sec("retreat_time"),  # seconds
        ),

    Pad(8),
    Struct("panic",
        from_to_sec("cowering_time"),  # seconds
        float_zero_to_one("friend_killed_panic_chance"),
        SEnum16("leader_type", *actor_types),

        Pad(2),
        float_zero_to_one("leader_killed_panic_chance"),
        float_zero_to_one("panic_damage_threshold"),
        float_wu("surprise_distance"),  # world units
        ),

    Pad(28),
    Struct("defensive",
        from_to_sec("hide_behind_cover_time"),  # seconds
        float_sec("hide_target_not_visible_time",
                  UNIT_SCALE=sec_unit_scale),  # seconds
        float_zero_to_one("hide_shield_fraction"),
        float_zero_to_one("attack_shield_fraction"),
        float_zero_to_one("pursue_shield_fraction"),

        Pad(16),
        SEnum16("defensive_crouch_type",
            "never",
            "danger",
            "low_shields",
            "hide_behind_shield",
            "any_target",
            "flood_shamble"
            ),

        Pad(2),
        Float("attacking_crouch_threshold"),
        Float("defending_crouch_threshold"),
        float_sec("mim_stand_time",  UNIT_SCALE=sec_unit_scale),  # seconds
        float_sec("mim_crouch_time", UNIT_SCALE=sec_unit_scale),  # seconds
        Float("defending_hide_time_modifier"),
        Float("attacking_evasion_threshold"),
        Float("defending_evasion_threshold"),
        float_zero_to_one("evasion_seek_cover_chance"),
        float_sec("evasion_delay_time", UNIT_SCALE=sec_unit_scale),  # seconds
        float_wu("max_seek_cover_distance"),  # world units
        float_zero_to_one("cover_damage_threshold"),
        float_sec("stalking_discovery_time",
                  UNIT_SCALE=sec_unit_scale),  # seconds
        float_wu("stalking_max_distance"),  # world units
        float_rad("stationary_facing_angle"),  # radians
        float_sec("change_facing_stand_time",
                  UNIT_SCALE=sec_unit_scale),  # seconds
        ),

    Pad(4),
    Struct("pursuit",
        from_to_sec("uncover_delay_time"),  # seconds
        from_to_sec("target_search_time"),  # seconds
        from_to_sec("pursuit_position_time"),  # seconds
        SInt16("coordinated_position_count", MIN=0),
        SInt16("normal_position_count", MIN=0),
        ),

    Pad(32),
    QStruct("berserk",
        float_sec("melee_attack_delay", UNIT_SCALE=sec_unit_scale),  # seconds
        float_wu("melee_fudge_factor"),  # world units
        float_sec("melee_charge_time",  UNIT_SCALE=sec_unit_scale),  # seconds
        float_wu("melee_leap_range_lower_bound"),  # world units
        float_wu("melee_leap_range_upper_bound"),  # world units
        Float("melee_leap_velocity", SIDETIP="world units/tick",
               UNIT_SCALE=per_sec_unit_scale),  # world units/tick
        float_zero_to_one("melee_leap_chance"),
        float_zero_to_one("melee_leap_ballistic"),
        float_zero_to_one("berserk_damage_amount"),
        float_zero_to_one("berserk_damage_threshold"),
        float_wu("berserk_proximity"),  # world units
        float_wu("suicide_sensing_dist"),  # world units
        float_zero_to_one("berserk_grenade_chance"),
        ),

    Pad(12),
    Struct("firing_positions",
        from_to_sec("guard_position_time"),  # seconds
        from_to_sec("combat_position_time"),  # seconds
        Float("old_position_avoid_dist"),  # world units
        Float("friend_avoid_dist"),  # world units
        ),

    Pad(40),
    Struct("communication",
        from_to_sec("noncombat_idle_speech_time"),  # seconds
        from_to_sec("combat_idle_speech_time"),  # seconds

        Pad(176),
        dependency("DO_NOT_USE_3", "actr"),
        ),
    SIZE=1272
    )


def get():
    return actr_def

actr_def = TagDef("actr",
    blam_header('actr', 2),
    actr_body,

    ext=".actor", endian=">", tag_cls=ActrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


pixl_unknown_1 = Struct("unknown_1",
    SInt16("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=2
    )


pixl_pixel_shader_unknown_2 = Struct("unknown_2",
    h3_string_id("unknown_0", VISIBLE=False),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=8
    )


pixl_pixel_shader = Struct("pixel_shader",
    h3_rawdata_ref("unknown_0", VISIBLE=False),
    h3_rawdata_ref("unknown_1", VISIBLE=False),
    h3_reflexive("unknown_2", pixl_pixel_shader_unknown_2),
    BytesRaw("unknown_3", SIZE=20, VISIBLE=False),
    SInt32("unknown_4", VISIBLE=False),
    UInt32("pixel_shader"),
    ENDIAN=">", SIZE=80
    )


pixl_body = Struct("tagdata",
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    h3_reflexive("unknown_1", pixl_unknown_1),
    BytesRaw("unknown_2", SIZE=4, VISIBLE=False),
    h3_reflexive("pixel_shaders", pixl_pixel_shader),
    ENDIAN=">", SIZE=32
    )


def get():
    return pixl_def

pixl_def = TagDef("pixl",
    h3_blam_header('pixl'),
    pixl_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["pixl"], endian=">", tag_cls=H3Tag
    )
